# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import json

from knack.util import CLIError

from azure.cli.core.commands.parameters import get_one_of_subscription_locations
from azure.cli.core.commands.arm import resource_exists

from ._client_factory import _compute_client_factory


def _resource_not_exists(cli_ctx, resource_type):
    def _handle_resource_not_exists(namespace):
        # TODO: hook up namespace._subscription_id once we support it
        ns, t = resource_type.split('/')
        if resource_exists(cli_ctx, namespace.resource_group_name, namespace.name, ns, t):
            raise CLIError('Resource {} of type {} in group {} already exists.'.format(
                namespace.name,
                resource_type,
                namespace.resource_group_name))
    return _handle_resource_not_exists


def _get_thread_count():
    return 5  # don't increase too much till https://github.com/Azure/msrestazure-for-python/issues/6 is fixed


def load_images_thru_services(cli_ctx, publisher, offer, sku, location):
    from concurrent.futures import ThreadPoolExecutor, as_completed
    all_images = []
    client = _compute_client_factory(cli_ctx)
    if location is None:
        location = get_one_of_subscription_locations(cli_ctx)

    def _load_images_from_publisher(publisher):
        offers = client.virtual_machine_images.list_offers(location, publisher)
        if offer:
            offers = [o for o in offers if _partial_matched(offer, o.name)]
        for o in offers:
            skus = client.virtual_machine_images.list_skus(location, publisher, o.name)
            if sku:
                skus = [s for s in skus if _partial_matched(sku, s.name)]
            for s in skus:
                images = client.virtual_machine_images.list(location, publisher, o.name, s.name)
                for i in images:
                    all_images.append({
                        'publisher': publisher,
                        'offer': o.name,
                        'sku': s.name,
                        'version': i.name})

    publishers = client.virtual_machine_images.list_publishers(location)
    if publisher:
        publishers = [p for p in publishers if _partial_matched(publisher, p.name)]

    publisher_num = len(publishers)
    if publisher_num > 1:
        with ThreadPoolExecutor(max_workers=_get_thread_count()) as executor:
            tasks = [executor.submit(_load_images_from_publisher, p.name) for p in publishers]
            for t in as_completed(tasks):
                t.result()  # don't use the result but expose exceptions from the threads
    elif publisher_num == 1:
        _load_images_from_publisher(publishers[0].name)

    return all_images


def load_images_from_aliases_doc(cli_ctx, publisher=None, offer=None, sku=None):
    import requests
    from azure.cli.core.cloud import CloudEndpointNotSetException
    from azure.cli.core.util import should_disable_connection_verify
    try:
        target_url = cli_ctx.cloud.endpoints.vm_image_alias_doc
    except CloudEndpointNotSetException:
        raise CLIError("'endpoint_vm_image_alias_doc' isn't configured. Please invoke 'az cloud update' to configure "
                       "it or use '--all' to retrieve images from server")
    # under hack mode(say through proxies with unsigned cert), opt out the cert verification
    response = requests.get(target_url, verify=(not should_disable_connection_verify()))
    if response.status_code != 200:
        raise CLIError("Failed to retrieve image alias doc '{}'. Error: '{}'".format(target_url, response))
    dic = json.loads(response.content.decode())
    try:
        all_images = []
        result = (dic['outputs']['aliases']['value'])
        for v in result.values():  # loop around os
            for alias, vv in v.items():  # loop around distros
                all_images.append({
                    'urnAlias': alias,
                    'publisher': vv['publisher'],
                    'offer': vv['offer'],
                    'sku': vv['sku'],
                    'version': vv['version']
                })

        all_images = [i for i in all_images if (_partial_matched(publisher, i['publisher']) and
                                                _partial_matched(offer, i['offer']) and
                                                _partial_matched(sku, i['sku']))]
        return all_images
    except KeyError:
        raise CLIError('Could not retrieve image list from {}'.format(target_url))


def load_extension_images_thru_services(cli_ctx, publisher, name, version, location, show_latest=False):
    from concurrent.futures import ThreadPoolExecutor, as_completed
    from distutils.version import LooseVersion  # pylint: disable=no-name-in-module,import-error
    all_images = []
    client = _compute_client_factory(cli_ctx)
    if location is None:
        location = get_one_of_subscription_locations(cli_ctx)

    def _load_extension_images_from_publisher(publisher):
        types = client.virtual_machine_extension_images.list_types(location, publisher)
        if name:
            types = [t for t in types if _partial_matched(name, t.name)]
        for t in types:
            versions = client.virtual_machine_extension_images.list_versions(location,
                                                                             publisher,
                                                                             t.name)
            if version:
                versions = [v for v in versions if _partial_matched(version, v.name)]

            if show_latest:
                # pylint: disable=no-member
                versions.sort(key=lambda v: LooseVersion(v.name), reverse=True)
                all_images.append({
                    'publisher': publisher,
                    'name': t.name,
                    'version': versions[0].name})
            else:
                for v in versions:
                    all_images.append({
                        'publisher': publisher,
                        'name': t.name,
                        'version': v.name})

    publishers = client.virtual_machine_images.list_publishers(location)
    if publisher:
        publishers = [p for p in publishers if _partial_matched(publisher, p.name)]

    publisher_num = len(publishers)
    if publisher_num > 1:
        with ThreadPoolExecutor(max_workers=_get_thread_count()) as executor:
            tasks = [executor.submit(_load_extension_images_from_publisher,
                                     p.name) for p in publishers]
            for t in as_completed(tasks):
                t.result()  # don't use the result but expose exceptions from the threads
    elif publisher_num == 1:
        _load_extension_images_from_publisher(publishers[0].name)

    return all_images


def get_vm_sizes(cli_ctx, location):
    return list(_compute_client_factory(cli_ctx).virtual_machine_sizes.list(location))


def _partial_matched(pattern, string):
    if not pattern:
        return True  # empty pattern means wildcard-match
    pattern = r'.*' + pattern
    import re
    return re.match(pattern, string, re.I)  # pylint: disable=no-member


def _create_image_instance(publisher, offer, sku, version):
    return {
        'publisher': publisher,
        'offer': offer,
        'sku': sku,
        'version': version
    }


def _get_latest_image_version(cli_ctx, location, publisher, offer, sku):
    top_one = _compute_client_factory(cli_ctx).virtual_machine_images.list(location,
                                                                           publisher,
                                                                           offer,
                                                                           sku,
                                                                           top=1,
                                                                           orderby='name desc')
    if not top_one:
        raise CLIError("Can't resolve the vesion of '{}:{}:{}'".format(publisher, offer, sku))
    return top_one[0].name



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

# pylint: disable=line-too-long, too-many-lines

from knack.help_files import helps

vm_ids_example = """        - name: {0}
          text: >
            az {1} --ids $(az vm list -g MyResourceGroup --query "[].id" -o tsv)
"""

helps['vm secret'] = """
    type: group
    short-summary: Manage VM secrets.
"""

helps['vm secret add'] = """
    type: command
    short-summary: Add a secret to a VM.
"""

helps['vm secret list'] = """
    type: command
    short-summary: List secrets on a VM.
"""

helps['vm secret remove'] = """
    type: command
    short-summary: Remove a secret from a VM.
"""

helps['vm secret format'] = """
    type: command
    short-summary: Transform secrets into a form that can be used by VMs and VMSSes.
    parameters:
        - name: --secrets -s
          long-summary: >
            The command will attempt to resolve the vault ID for each secret. If it is unable to do so,
            specify the vault ID to use for *all* secrets using: --keyvault NAME --resource-group NAME | --keyvault ID.
    examples:
        - name: Create a self-signed certificate with the default policy, and add it to a virtual machine.
          text: >
            az keyvault certificate create --vault-name vaultname -n cert1 \\
              -p "$(az keyvault certificate get-default-policy)"

            secrets=$(az keyvault secret list-versions --vault-name vaultname \\
              -n cert1 --query "[?attributes.enabled].id" -o tsv)

            vm_secrets=$(az vm secret format -s "$secrets")

            az vm create -g group-name -n vm-name --admin-username deploy  \\
              --image debian --secrets "$vm_secrets"
"""

helps['vm create'] = """
    type: command
    short-summary: Create an Azure Virtual Machine.
    long-summary: 'For an end-to-end tutorial, see https://docs.microsoft.com/azure/virtual-machines/virtual-machines-linux-quick-create-cli.'
    parameters:
        - name: --image
          type: string
          short-summary: >
            The name of the operating system image as a URN alias, URN, custom image name or ID, or VHD blob URI.
            This parameter is required unless using `--attach-os-disk.`
          populator-commands:
          - az vm image list
          - az vm image show
        - name: --ssh-key-value
          short-summary: The SSH public key or public key file path.
    examples:
        - name: Create a default Ubuntu VM with automatic SSH authentication.
          text: >
            az vm create -n MyVm -g MyResourceGroup --image UbuntuLTS
        - name: Create a default Windows Server VM with a private IP address.
          text: >
            az vm create -n MyVm -g MyResourceGroup --public-ip-address "" --image Win2012R2Datacenter
        - name: Create a VM from a custom managed image.
          text: >
            az vm create -g MyResourceGroup -n MyVm --image MyImage
        - name: Create a VM by attaching to a managed operating system disk.
          text: >
            az vm create -g MyResourceGroup -n MyVm --attach-os-disk MyOsDisk --os-type linux
        - name: 'Create an Ubuntu Linux VM using a cloud-init script for configuration. See: https://docs.microsoft.com/azure/virtual-machines/virtual-machines-linux-using-cloud-init.'
          text: >
            az vm create -g MyResourceGroup -n MyVm --image debian --custom-data MyCloudInitScript.yml
        - name: Create a Debian VM with SSH key authentication and a public DNS entry, located on an existing virtual network and availability set.
          text: |
            az vm create -n MyVm -g MyResourceGroup --image debian --vnet-name MyVnet --subnet subnet1 \\
                --availability-set MyAvailabilitySet --public-ip-address-dns-name MyUniqueDnsName \\
                --ssh-key-value @key-file
        - name: Create a simple Ubuntu Linux VM with a public IP address, DNS entry, two data disks (10GB and 20GB), and then generate ssh key pairs.
          text: |
            az vm create -n MyVm -g MyResourceGroup --public-ip-address-dns-name MyUniqueDnsName \\
                --image ubuntults --data-disk-sizes-gb 10 20 --size Standard_DS2_v2 \\
                --generate-ssh-keys
        - name: Create a Debian VM using Key Vault secrets.
          text: >
            az keyvault certificate create --vault-name vaultname -n cert1 \\
              -p "$(az keyvault certificate get-default-policy)"

            secrets=$(az keyvault secret list-versions --vault-name vaultname \\
              -n cert1 --query "[?attributes.enabled].id" -o tsv)

            vm_secrets=$(az vm secret format -s "$secrets") \n

            az vm create -g group-name -n vm-name --admin-username deploy  \\
              --image debian --secrets "$vm_secrets"
        - name: Create a CentOS VM with a system assigned identity. The VM will have a 'Contributor' role with access to a storage account.
          text: >
             az vm create -n MyVm -g rg1 --image centos --assign-identity --scope /subscriptions/99999999-1bf0-4dda-aec3-cb9272f09590/MyResourceGroup/myRG/providers/Microsoft.Storage/storageAccounts/storage1
        - name: Create a debian VM with a user assigned identity.
          text: >
             az vm create -n MyVm -g rg1 --image debian --assign-identity  /subscriptions/99999999-1bf0-4dda-aec3-cb9272f09590/resourcegroups/myRG/providers/Microsoft.ManagedIdentity/userAssignedIdentities/myID
        - name: Create a debian VM with both system and user assigned identity.
          text: >
             az vm create -n MyVm -g rg1 --image debian --assign-identity  [system] /subscriptions/99999999-1bf0-4dda-aec3-cb9272f09590/resourcegroups/myRG/providers/Microsoft.ManagedIdentity/userAssignedIdentities/myID
        - name: Create a VM in an availability zone in the current resource group's region
          min_profile: latest
          text: >
             az vm create -n MyVm -g MyResourceGroup --image Centos --zone 1
"""

helps['vmss create'] = """
    type: command
    short-summary: Create an Azure Virtual Machine Scale Set.
    long-summary: 'For an end-to-end tutorial, see https://docs.microsoft.com/azure/virtual-machine-scale-sets/virtual-machine-scale-sets-linux-create-cli.'
    parameters:
        - name: --image
          type: string
          short-summary: The name of the operating system image as a URN alias, URN, or URI.
    examples:
        - name: Create a Windows VM scale set with 5 instances, a load balancer, a public IP address, and a 2GB data disk.
          text: >
            az vmss create -n MyVmss -g MyResourceGroup --instance-count 5 --image Win2012R2Datacenter --data-disk-sizes-gb 2
        - name: Create a Linux VM scale set with an auto-generated ssh key pair, a public IP address, a DNS entry, an existing load balancer, and an existing virtual network.
          text: |
            az vmss create -n MyVmss -g MyResourceGroup --dns-name-for-public-ip MyGloballyUniqueDnsName \\
                --load-balancer MyLoadBalancer --vnet-name MyVnet --subnet MySubnet --image UbuntuLTS \\
                --generate-ssh-keys
        - name: Create a Linux VM scale set from a custom image using the default existing public SSH key.
          text: >
            az vmss create -n MyVmss -g MyResourceGroup --image MyImage
        - name: Create a Linux VM scale set with a load balancer and custom DNS servers. Each VM has a public-ip address and a custom domain name.
          text: >
            az vmss create -n MyVmss -g MyResourceGroup --image centos \\
                --public-ip-per-vm --vm-domain-name myvmss --dns-servers 10.0.0.6 10.0.0.5
        - name: 'Create a Linux VM scale set using a cloud-init script for configuration. See: https://docs.microsoft.com/azure/virtual-machines/virtual-machines-linux-using-cloud-init'
          text: >
            az vmss create -g MyResourceGroup -n MyVmss --image debian --custom-data MyCloudInitScript.yml
        - name: Create a Debian VM scaleset using Key Vault secrets.
          text: >
            az keyvault certificate create --vault-name vaultname -n cert1 \\
              -p "$(az keyvault certificate get-default-policy)"

            secrets=$(az keyvault secret list-versions --vault-name vaultname \\
              -n cert1 --query "[?attributes.enabled].id" -o tsv)

            vm_secrets=$(az vm secret format -s "$secrets") \n

            az vmss create -g group-name -n vm-name --admin-username deploy  \\
              --image debian --secrets "$vm_secrets"
        - name: Create a VM scaleset with system assigned identity. The VM will have a 'Contributor' Role with access to a storage account.
          text: >
             az vm create -n MyVmss -g MyResourceGroup --image centos --assign-identity --scope /subscriptions/99999999-1bf0-4dda-aec3-cb9272f09590/MyResourceGroup/myRG/providers/Microsoft.Storage/storageAccounts/storage1
        - name: Create a debian VM scaleset with a user assigned identity.
          text: >
             az vmss create -n MyVmss -g rg1 --image debian --assign-identity  /subscriptions/99999999-1bf0-4dda-aec3-cb9272f09590/resourcegroups/myRG/providers/Microsoft.ManagedIdentity/userAssignedIdentities/myID
        - name: Create a debian VM scaleset with both system and user assigned identity.
          text: >
             az vmss create -n MyVmss -g rg1 --image debian --assign-identity  [system] /subscriptions/99999999-1bf0-4dda-aec3-cb9272f09590/resourcegroups/myRG/providers/Microsoft.ManagedIdentity/userAssignedIdentities/myID
        - name: Create a single zone VM scaleset in the current resource group's region
          min_profile: latest
          text: >
             az vmss create -n MyVmss -g MyResourceGroup --image Centos --zones 1
"""

helps['vm availability-set create'] = """
    type: command
    short-summary: Create an Azure Availability Set.
    long-summary: 'For more information, see https://docs.microsoft.com/azure/virtual-machines/virtual-machines-linux-manage-availability.'
    examples:
        - name: Create an availability set.
          text: az vm availability-set create -n MyAvSet -g MyResourceGroup --platform-fault-domain-count 2 --platform-update-domain-count 2
"""

helps['vm availability-set update'] = """
    type: command
    short-summary: Update an Azure Availability Set.
    examples:
        - name: Update an availability set.
          text: az vm availability-set update -n MyAvSet -g MyResourceGroup
        - name: Update an availability set tag.
          text: az vm availability-set update -n MyAvSet -g MyResourceGroup --set tags.foo=value
        - name: Remove an availability set tag.
          text: az vm availability-set update -n MyAvSet -g MyResourceGroup --remove tags.foo
"""

helps['vm availability-set convert'] = """
    type: command
    short-summary: Convert an Azure Availability Set to contain VMs with managed disks.
    examples:
        - name: Convert an availabiity set to use managed disks by name.
          text: vm availability-set convert -g MyResourceGroup -n MyAvSet
        - name: Convert an availability set to use managed disks by ID.
          text: >
            az vm availability-set convert --ids $(az vm availability-set list -g MyResourceGroup --query "[].id" -o tsv)
"""

helps['vm extension set'] = """
    type: command
    short-summary: Set extensions for a VM.
    long-summary: Get extension details from `az vm extension image list`.
    examples:
        - name: Add a user account to a Linux VM.
          text: |
            az vm extension set -n VMAccessForLinux --publisher Microsoft.OSTCExtensions --version 1.4 \\
                --vm-name MyVm --resource-group MyResourceGroup \\
                --protected-settings '{"username":"user1", "ssh_key":"ssh_rsa ..."}'
"""

helps['vm availability-set delete'] = """
    type: command
    short-summary: Delete an availability set.
    examples:
        - name: Delete an availability set.
          text: az vm availability-set delete -n MyAvSet -g MyResourceGroup
"""

helps['vm availability-set list'] = """
    type: command
    short-summary: List availability sets.
    examples:
        - name: List availability sets.
          text: az vm availability-set list -g MyResourceGroup
"""

helps['vm availability-set list-sizes'] = """
    type: command
    short-summary: List VM sizes for an availability set.
    examples:
        - name: List VM sizes for an availability set.
          text: az vm availability-set list-sizes -n MyAvSet -g MyResourceGroup
"""

helps['vm availability-set show'] = """
    type: command
    short-summary: Get information for an availability set.
    examples:
        - name: Get information about an availability set.
          text: az vm availability-set show -n MyAvSet -g MyResourceGroup
"""

helps['vm update'] = """
    type: command
    short-summary: Update the properties of a VM.
    long-summary: Update VM objects and properties using paths that correspond to 'az vm show'.
    examples:
        - name: Add or update a tag.
          text: az vm update -n name -g group --set tags.tagName=tagValue
        - name: Remove a tag.
          text: az vm update -n name -g group --remove tags.tagName
        - name: Set the primary NIC of a VM.
          text: az vm update -n name -g group --set networkProfile.networkInterfaces[1].primary=false networkProfile.networkInterfaces[0].primary=true
        - name: Add a new non-primary NIC to a VM.
          text: az vm update -n name -g group --add networkProfile.networkInterfaces primary=false id=<NIC_ID>
        - name: Remove the fourth NIC from a VM.
          text: az vm update -n name -g group --remove networkProfile.networkInterfaces 3
"""

helps['vmss deallocate'] = """
    type: command
    short-summary: Deallocate VMs within a VMSS.
"""

helps['vmss delete-instances'] = """
    type: command
    short-summary: Delete VMs within a VMSS.
"""

helps['vmss get-instance-view'] = """
    type: command
    short-summary: View an instance of a VMSS.
    parameters:
        - name: --instance-id
          short-summary: A VM instance ID or "*" to list instance view for all VMs in a scale set.

"""

helps['vmss list'] = """
    type: command
    short-summary: List VMSS.
"""

helps['vmss reimage'] = """
    type: command
    short-summary: Reimage VMs within a VMSS.
    parameters:
        - name: --instance-id
          short-summary: VM instance ID. If missing, reimage all instances.
"""

helps['vmss restart'] = """
    type: command
    short-summary: Restart VMs within a VMSS.
"""

helps['vmss scale'] = """
    type: command
    short-summary: Change the number of VMs within a VMSS.
    parameters:
        - name: --new-capacity
          short-summary: Number of VMs in the VMSS.
"""

helps['vmss show'] = """
    type: command
    short-summary: Get details on VMs within a VMSS.
    parameters:
        - name: --instance-id
          short-summary: VM instance ID. If missing, show the VMSS.
"""

helps['vmss start'] = """
    type: command
    short-summary: Start VMs within a VMSS.
"""

helps['vmss stop'] = """
    type: command
    short-summary: Power off (stop) VMs within a VMSS.
"""

helps['vmss update'] = """
    type: command
    short-summary: Update a VMSS.
"""

helps['vmss update-instances'] = """
    type: command
    short-summary: Upgrade VMs within a VMSS.
"""

helps['vmss wait'] = """
    type: command
    short-summary: Place the CLI in a waiting state until a condition of a scale set is met.
"""

helps['vmss disk'] = """
    type: group
    short-summary: Manage data disks of a VMSS.
"""

helps['vmss disk attach'] = """
    type: command
    short-summary: Attach managed data disks to a scale set or its instances.
"""

helps['vmss disk detach'] = """
    type: command
    short-summary: Detach managed data disks from a scale set or its instances.
"""

helps['vmss nic'] = """
    type: group
    short-summary: Manage network interfaces of a VMSS.
"""

helps['vmss rolling-upgrade'] = """
    type: group
    short-summary: (PREVIEW) Manage rolling upgrades.
"""

helps['vm convert'] = """
    type: command
    short-summary: Convert a VM with unmanaged disks to use managed disks.
    examples:
        - name: Convert a VM with unmanaged disks to use managed disks.
          text: az vm convert -g MyResourceGroup -n MyVm
{0}
""".format(vm_ids_example.format('Convert all VMs with unmanaged disks in a resource group to use managed disks.', 'vm convert'))

helps['vm'] = """
    type: group
    short-summary: Provision Linux or Windows virtual machines.
"""
helps['vm user'] = """
    type: group
    short-summary: Manage user accounts for a VM.
"""

helps['vm user delete'] = """
    type: command
    short-summary: Delete a user account from a VM.
    examples:
        - name: Delete a user account.
          text: az vm user delete -u username -n MyVm -g MyResourceGroup
{0}
""".format(vm_ids_example.format('Delete a user on all VMs in a resource group.', 'az vm user delete -u username'))

helps['vm user reset-ssh'] = """
    type: command
    short-summary: Reset the SSH configuration on a VM.
    long-summary: >
        The extension will restart the SSH service, open the SSH port on your VM, and reset the SSH configuration to default values. The user account (name, password, and SSH keys) are not changed.
    examples:
        - name: Reset the SSH configuration.
          text: az vm user reset-ssh -n MyVm -g MyResourceGroup
{0}
""".format(vm_ids_example.format("Reset the SSH server on all VMs in a resource group.", 'vm user reset-ssh'))

helps['vm user update'] = """
    type: command
    short-summary: Update a user account.
    parameters:
        - name: --ssh-key-value
          short-summary: SSH public key file value or public key file path
    examples:
        - name: Update a Windows user account.
          text: az vm user update -u username -p password -n MyVm -g MyResourceGroup
        - name: Update a Linux user account.
          text: az vm user update -u username --ssh-key-value "$(< ~/.ssh/id_rsa.pub)" -n MyVm -g MyResourceGroup
{0}
""".format(vm_ids_example.format('Update a user on all VMs in a resource group.',
                                 'vm user update -u username --ssh-key-value "$(< ~/.ssh/id_rsa.pub)"'))

helps['vm availability-set'] = """
    type: group
    short-summary: Group resources into availability sets.
    long-summary: >
        To provide redundancy to an application, it is recommended to group two or more virtual machines in an availability set.
        This configuration ensures that during either a planned or unplanned maintenance event, at least one virtual machine
        will be available.
"""

helps['vm boot-diagnostics'] = """
    type: group
    short-summary: Troubleshoot the startup of an Azure Virtual Machine.
    long-summary: Use this feature to troubleshoot boot failures for custom or platform images.
"""

vm_boot_diagnostics_disable = 'vm boot-diagnostics disable'
helps[vm_boot_diagnostics_disable] = """
    type: command
    short-summary: Disable the boot diagnostics on a VM.
    examples:
{0}
""".format(vm_ids_example.format('Disable boot diagnostics on all VMs in a resource group.', vm_boot_diagnostics_disable))

vm_boot_diagnostics_enable = 'vm boot-diagnostics enable'
vm_boot_diagnostics_enable_cmd = "{0} --storage https://mystor.blob.core.windows.net/".format(vm_boot_diagnostics_enable)
helps[vm_boot_diagnostics_enable] = """
    type: command
    short-summary: Enable the boot diagnostics on a VM.
    parameters:
        - name: --storage
          short-summary: Name or URI of a storage account (e.g. https://your_storage_account_name.blob.core.windows.net/)
    examples:
{0}
""".format(vm_ids_example.format('Enable boot diagnostics on all VMs in a resource group.', vm_boot_diagnostics_enable_cmd))

boot_diagnostics_log = 'vm boot-diagnostics get-boot-log'
helps[boot_diagnostics_log] = """
    type: command
    short-summary: Get the boot diagnostics log from a VM.
    examples:
{0}
""".format(vm_ids_example.format('Get diagnostics logs for all VMs in a resource group.', boot_diagnostics_log))

helps['acs'] = """
    type: group
    short-summary: Manage Azure Container Services.
"""

helps['acs create'] = """
    type: command
    short-summary: Create a container service.
    examples:
        - name: Create a Kubernetes container service and generate SSH keys to connect to it.
          text: >
            az acs create -g MyResourceGroup -n MyContainerService --orchestrator-type kubernetes --generate-ssh-keys
"""

helps['acs delete'] = """
    type: command
    short-summary: Delete a container service.
"""

helps['acs list'] = """
    type: command
    short-summary: List container services.
"""

helps['acs show'] = """
    type: command
    short-summary: Get the details for a container service.
"""

helps['acs scale'] = """
    type: command
    short-summary: Change the private agent count of a container service.
"""

helps['vm diagnostics'] = """
    type: group
    short-summary: Configure the Azure Virtual Machine diagnostics extension.
"""

helps['vm diagnostics get-default-config'] = """
    type: command
    short-summary: Get the default configuration settings for a VM.
    examples:
        - name: Get the default diagnostics for a Linux VM and override the storage account name and the VM resource ID.
          text: |
            az vm diagnostics get-default-config \\
                | sed "s#__DIAGNOSTIC_STORAGE_ACCOUNT__#MyStorageAccount#g" \\
                | sed "s#__VM_OR_VMSS_RESOURCE_ID__#MyVmResourceId#g"
        - name: Get the default diagnostics for a Windows VM.
          text: >
            az vm diagnostics get-default-config --is-windows-os
"""

helps['vm diagnostics set'] = """
    type: command
    short-summary: Configure the Azure VM diagnostics extension.
    examples:
        - name: Set up default diagnostics on a Linux VM for Azure Portal VM metrics graphs and syslog collection.
          text: >
            # Set the following 3 parameters first.\n
            my_resource_group=<Resource group name containing your Linux VM and the storage account>\n
            my_linux_vm=<Your Azure Linux VM name>\n
            my_diagnostic_storage_account=<Your Azure storage account for storing VM diagnostic data>\n

            my_vm_resource_id=$(az vm show -g $my_resource_group -n $my_linux_vm --query "id" -o tsv)\n

            default_config=$(az vm diagnostics get-default-config \\
                | sed "s#__DIAGNOSTIC_STORAGE_ACCOUNT__#$my_diagnostic_storage_account#g" \\
                | sed "s#__VM_OR_VMSS_RESOURCE_ID__#$my_vm_resource_id#g")

            storage_sastoken=$(az storage account generate-sas \\
                --account-name $my_diagnostic_storage_account --expiry 9999-12-31T23:59Z \\
                --permissions wlacu --resource-types co --services bt -o tsv)

            protected_settings="{'storageAccountName': '{my_diagnostic_storage_account}', \\
                'storageAccountSasToken': '{storage_sastoken}'}"

            az vm diagnostics set --settings "{default_config}" \\
                --protected-settings "{protected_settings}" \\
                --resource-group $my_resource_group --vm-name $my_linux_vm
"""

disk_long_summary = """
        Just like any other computer, virtual machines in Azure use disks as a place to store an operating system, applications, and data.
        All Azure virtual machines have at least two disks: An operating system disk, and a temporary disk.
        The operating system disk is created from an image, and both the operating system disk and the image are actually virtual hard disks (VHDs)
        stored in an Azure storage account. Virtual machines also can have one or more data disks, that are also stored as VHDs.

        Operating System Disk
        Every virtual machine has one attached operating system disk. It's registered as a SATA drive and is labeled /dev/sda by default.
        This disk has a maximum capacity of 1023 gigabytes (GB).

        Temporary disk
        The temporary disk is automatically created for you. On Linux virtual machines, the disk is typically /dev/sdb and is formatted and
        mounted to /mnt/resource by the Azure Linux Agent. The size of the temporary disk varies, based on the size of the virtual machine.

        Data disk
        A data disk is a VHD that's attached to a virtual machine to store application data, or other data you need to keep. Data disks are
        registered as SCSI drives and are labeled by the creator. Each data disk has a maximum capacity of 1023 GB. The size of the virtual
        machine determines how many data disks can be attached and the type of storage that can be used to host the disks.
"""

helps['vm disk'] = """
    type: group
    short-summary: Manage the managed data disks attached to a VM.
    long-summary: >
{0}
""".format(disk_long_summary)

helps['vm unmanaged-disk'] = """
    type: group
    short-summary: Manage the unmanaged data disks attached to a VM.
    long-summary: >
{0}
""".format(disk_long_summary)

helps['vm unmanaged-disk attach'] = """
    type: command
    short-summary: Attach an unmanaged persistent disk to a VM.
    long-summary: This allows for the preservation of data, even if the VM is reprovisioned due to maintenance or resizing.
    examples:
        - name: Attach a new default sized (1023 GB) unmanaged data disk to a VM.
          text: az vm unmanaged-disk attach -g MyResourceGroup --vm-name MyVm
        - name: Attach an existing data disk to a VM as unmanaged.
          text: >
            az vm unmanaged-disk attach -g MyResourceGroup --vm-name MyVm \\
                --vhd-uri https://mystorage.blob.core.windows.net/vhds/d1.vhd
"""

helps['vm unmanaged-disk detach'] = """
    type: command
    short-summary: Detatch an unmanaged disk from a VM.
    examples:
        - name: Detach a data disk from a VM.
          text: >
            az vm unmanaged-disk detach -g MyResourceGroup --vm-name MyVm -n disk_name
"""

helps['vm unmanaged-disk list'] = """
    type: command
    short-summary: List unamanaged disks of a VM.
    examples:
        - name: List the unmanaged disks attached to a VM.
          text: az vm unmanaged-disk list -g MyResourceGroup --vm-name MyVm
        - name: List unamanaged disks with IDs containing the string "data_disk".
          text: >
            az vm unmanaged-disk list --ids \\
                $(az resource list --query "[?contains(name, 'data_disk')].id" -o tsv)
"""

helps['vm disk detach'] = """
    type: command
    short-summary: Detatch a managed disk from a VM.
    examples:
        - name: Detach a data disk from a VM.
          text: >
            az vm disk detach -g MyResourceGroup --vm-name MyVm -n disk_name
"""

helps['vm disk attach'] = """
    type: command
    short-summary: Attach a managed persistent disk to a VM.
    long-summary: This allows for the preservation of data, even if the VM is reprovisioned due to maintenance or resizing.
    examples:
        - name: Attach a new default sized (1023 GB) managed data disk to a VM.
          text: az vm disk attach -g MyResourceGroup --vm-name MyVm --disk disk_name --new
"""

helps['vm encryption'] = """
    type: group
    short-summary: Manage encryption of VM disks.
"""

helps['vm encryption enable'] = """
    type: command
    short-summary: Enable disk encryption on the OS disk and/or data disks.
    parameters:
        - name: --aad-client-id
          short-summary: Client ID of an AAD app with permissions to write secrets to the key vault.
        - name: --aad-client-secret
          short-summary: Client secret of the AAD app with permissions to write secrets to the key vault.
        - name: --aad-client-cert-thumbprint
          short-summary: Thumbprint of the AAD app certificate with permissions to write secrets to the key vault.
"""

helps['vm encryption disable'] = """
    type: command
    short-summary: Disable disk encryption on the OS disk and/or data disks.
"""

helps['vm encryption show'] = """
    type: command
    short-summary: Show encryption status.
"""

helps['vm extension'] = """
    type: group
    short-summary: Manage extensions on VMs.
    long-summary: >
        Extensions are small applications that provide post-deployment configuration and automation tasks on Azure virtual machines.
        For example, if a virtual machine requires software installation, anti-virus protection, or Docker configuration, a VM extension
        can be used to complete these tasks. Extensions can be bundled with a new virtual machine deployment or run against any existing system.
"""

helps['vm extension list'] = """
    type: command
    short-summary:  List the extensions attached to a VM.
    examples:
        - name: List attached extensions to a named VM.
          text: az vm extension list -g MyResourceGroup --vm-name MyVm
        - name: List attached extensions with IDs containing the string "MyExtension".
          text: >
            az vm extension list --ids \\
                $(az resource list --query "[?contains(name, 'MyExtension')].id" -o tsv)
"""

helps['vm extension delete'] = """
    type: command
    short-summary: Remove an extension attached to a VM.
    examples:
        - name: Use a VM name and extension to delete an extension from a VM.
          text: az vm extension delete -g MyResourceGroup --vm-name MyVm -n extension_name
        - name: Delete extensions with IDs containing the string "MyExtension" from a VM.
          text: >
            az vm extension delete --ids \\
                $(az resource list --query "[?contains(name, 'MyExtension')].id" -o tsv)
"""

helps['vm extension show'] = """
    type: command
    short-summary: Display information about extensions attached to a VM.
    examples:
        - name: Use VM name and extension name to show the extensions attached to a VM.
          text: az vm extension show -g MyResourceGroup --vm-name MyVm -n extension_name
"""

helps['vm extension image'] = """
    type: group
    short-summary: Find the available VM extensions for a subscription and region.
"""

helps['vm extension image list'] = """
    type: command
    short-summary: List the information on available extensions.
    examples:
        - name: List the unique publishers for extensions.
          text: az vm extension image list --query "[].publisher" -o tsv | sort -u
        - name: Find extensions with "Docker" in the name.
          text: az vm extension image list --query "[].name" -o tsv | sort -u | grep Docker
        - name: List extension names where the publisher name starts with "Microsoft.Azure.App".
          text: |
            az vm extension image list --query \\
                "[?starts_with(publisher, 'Microsoft.Azure.App')].publisher" \\
                -o tsv | sort -u | xargs -I{} az vm extension image list-names --publisher {} -l westus
"""

helps['vm extension image list-names'] = """
    type: command
    short-summary: List the names of available extensions.
    examples:
        - name: Find Docker extensions by publisher and location.
          text: >
            az vm extension image list-names --publisher Microsoft.Azure.Extensions \\
                -l westus --query "[?starts_with(name, 'Docker')]"
        - name: Find CustomScript extensions by publisher and location.
          text: >
            az vm extension image list-names --publisher Microsoft.Azure.Extensions \\
                -l westus --query "[?starts_with(name, 'Custom')]"
"""

helps['vm extension image list-versions'] = """
    type: command
    short-summary: List the versions for available extensions.
    examples:
        - name: Find the available versions for the Docker extension.
          text: >
            az vm extension image list-versions --publisher Microsoft.Azure.Extensions \\
                -l westus -n DockerExtension -otable
"""

helps['vm extension image show'] = """
    type: command
    short-summary: Display information for an extension.
    examples:
        - name: Show the CustomScript extension version 2.0.2.
          text: >
            az vm extension image show -l westus -n CustomScript \\
              --publisher Microsoft.Azure.Extensions --version 2.0.2
        - name: Show the latest version of the Docker extension.
          text: >
            publisher=Microsoft.Azure.Extensions\n
            extension=DockerExtension\n
            location=westus\n

            latest=$(az vm extension image list-versions \\
              --publisher {publisher} -l {location} -n {extension} \\
              --query "[].name" -o tsv | sort | tail -n 1)

            az vm extension image show -l {location} \\
              --publisher {publisher} -n {extension} --version {latest}
"""

helps['vm image'] = """
    type: group
    short-summary: Information on available virtual machine images.
"""

helps['vm image list'] = """
    type: command
    short-summary: List the VM/VMSS images available in the Azure Marketplace.
    parameters:
        - name: --all
          short-summary: Retrieve image list from live Azure service rather using an offline image list
        - name: --offer -f
          short-summary: Image offer name, partial name is accepted
        - name: --publisher -p
          short-summary: Image publisher name, partial name is accepted
        - name: --sku -s
          short-summary: Image sku name, partial name is accepted
    examples:
        - name: List all available images.
          text: az vm image list --all
        - name: List all offline cached CentOS images.
          text: az vm image list -f CentOS
        - name: List all CentOS images.
          text: az vm image list -f CentOS --all
"""

helps['vm image list-offers'] = """
    type: command
    short-summary: List the VM image offers available in the Azure Marketplace.
    examples:
        - name: List all offers from Microsoft in the West US region.
          text: az vm image list-offers -l westus -p Microsoft
        - name: List all offers from OpenLocic in the West US region.
          text: az vm image list-offers -l westus -p OpenLogic
"""

helps['vm image list-publishers'] = """
    type: command
    short-summary: List the VM image publishers available in the Azure Marketplace.
    examples:
        - name: List all publishers in the West US region.
          text: az vm image list-publishers -l westus
        - name: List all publishers with names starting with "Open" in westus.
          text: az vm image list-publishers -l westus --query "[?starts_with(name, 'Open')]"
"""

helps['vm image list-skus'] = """
    type: command
    short-summary: List the VM image SKUs available in the Azure Marketplace.
    examples:
        - name: List all skus available for CentOS published by OpenLogic in the West US region.
          text: az vm image list-skus -l westus -f CentOS -p OpenLogic
"""

helps['vm image show'] = """
    type: command
    short-summary: Get the details for a VM image available in the Azure Marketplace.
    examples:
        - name: Show information for the latest available CentOS image from OpenLogic.
          text: >
            latest=$(az vm image list -p OpenLogic -s 7.3 --all --query \\
                "[?offer=='CentOS'].version" -o tsv | sort -u | tail -n 1)
            az vm image show -l westus -f CentOS -p OpenLogic --s 7.3 --version {latest}
"""

helps['vm image accept-terms'] = """
    type: command
    short-summary: Accept Azure Marketplace term so that the image can be used to create VMs
"""

helps['vm nic'] = """
    type: group
    short-summary: Manage network interfaces. See also `az network nic`.
    long-summary: >
        A network interface (NIC) is the interconnection between a VM and the underlying software
        network. For more information, see https://docs.microsoft.com/azure/virtual-network/virtual-network-network-interface-overview.
"""

helps['vm nic list'] = """
    type: command
    short-summary: List the NICs available on a VM.
    examples:
        - name: List all of the NICs on a VM.
          text: az vm nic list -g MyResourceGroup --vm-name MyVm
"""

helps['vm nic add'] = """
    type: command
    short-summary: Add existing NICs to a VM.
    examples:
        - name: Add two NICs to a VM.
          text: az vm nic add -g MyResourceGroup --vm-name MyVm --nics nic_name1 nic_name2
"""

helps['vm nic remove'] = """
    type: command
    short-summary: Remove NICs from a VM.
    examples:
        - name: Remove two NICs from a VM.
          text: az vm nic remove -g MyResourceGroup --vm-name MyVm --nics nic_name1 nic_name2
"""

helps['vm nic show'] = """
    type: command
    short-summary: Display information for a NIC attached to a VM.
    examples:
        - name: Show details of a NIC on a VM.
          text: az vm nic show -g MyResourceGroup --vm-name MyVm --nic nic_name1
"""

helps['vm nic set'] = """
    type: command
    short-summary: Configure settings of a NIC attached to a VM.
    examples:
        - name: Set a NIC on a VM to be the primary interface.
          text: az vm nic set -g MyResourceGroup --vm-name MyVm --nic nic_name1 nic_name2 --primary-nic nic_name2
"""

helps['vmss'] = """
    type: group
    short-summary: Manage groupings of virtual machines in an Azure Virtual Machine Scale Set (VMSS).
"""

helps['vmss diagnostics'] = """
    type: group
    short-summary: Configure the Azure Virtual Machine Scale Set diagnostics extension.
"""

helps['vmss diagnostics get-default-config'] = """
    type: command
    short-summary: Show the default config file which defines data to be collected.
"""

helps['vmss diagnostics set'] = """
    type: command
    short-summary: Enable diagnostics on a VMSS.
"""


helps['vmss list-instance-connection-info'] = """
    type: command
    short-summary: Get the IP address and port number used to connect to individual VM instances within a set.
"""

helps['vmss list-instance-public-ips'] = """
    type: command
    short-summary: List public IP addresses of VM instances within a set.
"""

helps['vmss extension'] = """
    type: group
    short-summary: Manage extensions on a VM scale set.
"""

helps['vmss extension delete'] = """
    type: command
    short-summary: Delete an extension from a VMSS.
"""

helps['vmss extension list'] = """
    type: command
    short-summary: List extensions associated with a VMSS.
"""

helps['vmss extension set'] = """
    type: command
    short-summary: Add an extension to a VMSS or update an existing extension.
    long-summary: Get extension details from `az vmss extension image list`.
"""

helps['vmss extension show'] = """
    type: command
    short-summary: Show details on a VMSS extension.
"""

helps['vmss extension image'] = """
    type: group
    short-summary: Find the available VM extensions for a subscription and region.
"""

helps['vmss extension image list'] = """
    type: command
    short-summary: List the information on available extensions.
    examples:
        - name: List the unique publishers for extensions.
          text: az vmss extension image list --query "[].publisher" -o tsv | sort -u
        - name: Find extensions with "Docker" in the name.
          text: az vmss extension image list --query "[].name" -o tsv | sort -u | grep Docker
        - name: List extension names where the publisher name starts with "Microsoft.Azure.App".
          text: |
            az vmss extension image list --query \\
                "[?starts_with(publisher, 'Microsoft.Azure.App')].publisher" \\
                -o tsv | sort -u | xargs -I{} az vmss extension image list-names --publisher {} -l westus
"""


deallocate_generalize_capture = """        - name: Deallocate, generalize, and capture a stopped virtual machine.
          text: |
            az vm deallocate -g MyResourceGroup -n MyVm
            az vm generalize -g MyResourceGroup -n MyVm
            az vm capture -g MyResourceGroup -n MyVm --vhd-name-prefix MyPrefix
        - name: Deallocate, generalize, and capture multiple stopped virtual machines.
          text: |
            vms_ids=$(az vm list -g MyResourceGroup --query "[].id" -o tsv)
            az vm deallocate --ids {vms_ids}
            az vm generalize --ids {vms_ids}
            az vm capture --ids {vms_ids} --vhd-name-prefix MyPrefix
"""

helps['vmss encryption'] = """
    type: group
    short-summary: (PREVIEW) Manage encryption of VMSS.
"""

helps['vmss encryption enable'] = """
    type: command
    short-summary: Encrypt a VMSS with managed disks.
    examples:
        - name: encrypt a VM scale set using a key vault in the same resource group
          text: >
            az vmss encryption enable -g MyResourceGroup -n MyVm --disk-encryption-keyvault myvault
"""

helps['vmss encryption disable'] = """
    type: command
    short-summary: Disable the encryption on a VMSS with managed disks.
    examples:
        - name: disable encryption a VMSS
          text: >
            az vmss encryption disable -g MyResourceGroup -n MyVm
"""

helps['vmss encryption show'] = """
    type: command
    short-summary: Show encryption status.
"""

helps['vm capture'] = """
    type: command
    short-summary: Capture information for a stopped VM.
    long-summary: 'For an end-to-end tutorial, see https://docs.microsoft.com/azure/virtual-machines/virtual-machines-linux-capture-image'
    parameters:
        - name: --vhd-name-prefix
          type: string
          short-summary: The VHD name prefix specify for the VM disks.
        - name: --storage-container
          short-summary: The storage account container name in which to save the disks.
        - name: --overwrite
          short-summary: Overwrite the existing disk file.
    examples:
{0}
""".format(deallocate_generalize_capture)

helps['vm delete'] = """
    type: command
    short-summary: Delete a VM.
    examples:
        - name: Delete a VM without a prompt for confirmation.
          text: >
            az vm delete -g MyResourceGroup -n MyVm --yes
{0}
""".format(vm_ids_example.format('Delete all VMs in a resource group.', 'vm delete'))

helps['vm deallocate'] = """
    type: command
    short-summary: Deallocate a VM.
    long-summary: 'For an end-to-end tutorial, see https://docs.microsoft.com/azure/virtual-machines/virtual-machines-linux-capture-image'
    examples:
{0}
""".format(deallocate_generalize_capture)

helps['vm generalize'] = """
    type: command
    short-summary: Mark a VM as generalized, allowing it to be imaged for multiple deployments.
    long-summary: 'For an end-to-end tutorial, see https://docs.microsoft.com/azure/virtual-machines/virtual-machines-linux-capture-image'
    examples:
{0}
""".format(deallocate_generalize_capture)

helps['vm get-instance-view'] = """
    type: command
    short-summary: Get instance information about a VM.
    examples:
        - name: Use a resource group and name to get instance view information of a VM.
          text: az vm get-instance-view -g MyResourceGroup -n MyVm
{0}
""".format(vm_ids_example.format('Get instance views for all VMs in a resource group.', 'vm get-instance-view'))

helps['vm list'] = """
    type: command
    short-summary: List details of Virtual Machines.
    long-summary: 'For more information on querying information about Virtual Machines, see https://docs.microsoft.com/en-us/cli/azure/query-az-cli2'
    examples:
        - name: List all VMs.
          text: az vm list
        - name: List all VMs by resource group.
          text: az vm list -g MyResourceGroup
        - name: List all VMs by resource group with details.
          text: az vm list -g MyResourceGroup -d
"""

helps['vm list-ip-addresses'] = """
    type: command
    short-summary: List IP addresses associated with a VM.
    examples:
        - name: Get the IP addresses for a VM.
          text: az vm list-ip-addresses -g MyResourceGroup -n MyVm
{0}
""".format(vm_ids_example.format('Get IP addresses for all VMs in a resource group.', 'vm list-ip-addresses'))

helps['vm list-sizes'] = """
    type: command
    short-summary: List available sizes for VMs.
    examples:
        - name: List the available VM sizes in the West US region.
          text: az vm list-sizes -l westus
"""

helps['vm list-usage'] = """
    type: command
    short-summary: List available usage resources for VMs.
    examples:
        - name: Get the compute resource usage for the West US region.
          text: az vm list-usage -l westus
"""

helps['vm list-vm-resize-options'] = """
    type: command
    short-summary: List available resizing options for VMs.
    examples:
        - name: List all available VM sizes for resizing.
          text: az vm list-vm-resize-options -g MyResourceGroup -n MyVm
{0}
""".format(vm_ids_example.format('List available sizes for all VMs in a resource group.', 'vm list-vm-resize-options'))

helps['vm list-skus'] = """
    type: command
    short-summary: Get details for compute-related resource SKUs.
    long-summary: This command incorporates subscription level restriction, offering the most accurate information.
    examples:
        - name: List all SKUs in the West US region.
          text: az vm list-skus -l westus
"""

helps['vm open-port'] = """
    type: command
    short-summary: Opens a VM to inbound traffic on specified ports.
    long-summary: >
        Adds a security rule to the network security group (NSG) that is attached to the VM's
        network interface (NIC) or subnet. The existing NSG will be used or a new one will be
        created. The rule name is 'open-port-{{port}}' and will overwrite an existing rule with
        this name. For multi-NIC VMs, or for more fine-grained control, use the appropriate
        network commands directly (nsg rule create, etc).
    examples:
        - name: Open all ports on a VM to inbound traffic.
          text: az vm open-port -g MyResourceGroup -n MyVm --port '*'
        - name: Open a range of ports on a VM to inbound traffic with the highest priority.
          text: az vm open-port -g MyResourceGroup -n MyVm --port 80-100 --priority 100
{0}
""".format(vm_ids_example.format('Open all ports for all VMs in a resource group.', 'vm open-port'))

helps['vm redeploy'] = """
    type: command
    short-summary: Redeploy an existing VM.
    examples:
        - name: Redeploy a VM.
          text: az vm redeploy -g MyResourceGroup -n MyVm
{0}
""".format(vm_ids_example.format('Redeploy all VMs in a resource group.', 'vm redeploy'))

helps['vm resize'] = """
    type: command
    short-summary: Update a VM's size.
    parameters:
        - name: --size
          type: string
          short-summary: The VM size.
          populator-commands:
          - az vm list-vm-resize-options
    examples:
        - name: Resize a VM.
          text: az vm resize -g MyResourceGroup -n MyVm --size Standard_DS3_v2
{0}
""".format(vm_ids_example.format('Resize all VMs in a resource group.', 'vm resize --size Standard_DS3_v2'))

helps['vm restart'] = """
    type: command
    short-summary: Restart VMs.
    examples:
        - name: Restart a VM.
          text: az vm restart -g MyResourceGroup -n MyVm
{0}
""".format(vm_ids_example.format('Restart all VMs in a resource group.', 'vm restart'))

helps['vm show'] = """
    type: command
    short-summary: Get the details of a VM.
    examples:
        - name: Show information about a VM.
          text: az vm show -g MyResourceGroup -n MyVm -d
{0}
""".format(vm_ids_example.format('Get the details for all VMs in a resource group.', 'vm show -d'))

helps['vm start'] = """
    type: command
    short-summary: Start a stopped VM.
    examples:
        - name: Start a stopped VM.
          text: az vm start -g MyResourceGroup -n MyVm
{0}
""".format(vm_ids_example.format('Start all VMs in a resource group.', 'vm start'))

helps['vm stop'] = """
    type: command
    short-summary: Stop a running VM.
    examples:
        - name: Stop a running VM.
          text: az vm stop -g MyResourceGroup -n MyVm
{0}
""".format(vm_ids_example.format('Stop all VMs in a resource group.', 'vm stop'))

helps['vm wait'] = """
    type: command
    short-summary: Place the CLI in a waiting state until a condition of the VM is met.
    examples:
        - name: Wait until a VM is created.
          text: az vm wait -g MyResourceGroup -n MyVm --created
{0}
""".format(vm_ids_example.format('Wait until all VMs in a resource group are deleted.', 'vm wait --deleted'))

helps['vm identity'] = """
    type: group
    short-summary: manage service identities of a VM
"""

helps['vm identity assign'] = """
    type: command
    short-summary: Enable managed service identity on a VM.
    long-summary: This is required to authenticate and interact with other Azure services using bearer tokens.
    examples:
        - name: Enable identity on a VM with the 'Reader' role.
          text: az vm identity assign -g MyResourceGroup -n MyVm --role Reader
"""

helps['vm identity remove'] = """
    type: command
    short-summary: (PREVIEW) Remove user assigned identities from a VM.
    examples:
        - name: Remove 2 identities which are in the same resource group with the VM
          text: az vm identity remove -g MyResourceGroup -n MyVm --identities readerId writerId
"""

helps['vm identity show'] = """
    type: command
    short-summary: display VM's managed identity info.
"""

helps['vm run-command'] = """
    type: group
    short-summary: Manage run commands on a Virtual Machine
"""

helps['vm run-command invoke'] = """
    type: command
    short-summary: run command on a vm
    examples:
        - name: install nginx on a vm
          text: az vm run-command invoke -g MyResourceGroup -n MyVm --command-id RunShellScript --scripts "sudo apt-get update && sudo apt-get install -y nginx"
        - name: invoke command with parameters
          text: az vm run-command invoke -g MyResourceGroup -n MyVm --command-id RunShellScript --scripts 'echo $0 $1' --parameters hello world
"""

helps['vmss identity'] = """
    type: group
    short-summary: manage service identities of a VM scaleset.
"""

helps['vmss identity assign'] = """
    type: command
    short-summary: Enable managed service identity on a VMSS.
    long-summary: This is required to authenticate and interact with other Azure services using bearer tokens.
    examples:
        - name: Enable identity on a VMSS with the 'Owner' role.
          text: az vmss identity assign -g MyResourceGroup -n MyVmss --role Owner
"""

helps['vmss identity remove'] = """
    type: command
    short-summary: (PREVIEW) Remove user assigned identities from a VM scaleset.
    examples:
        - name: Remove 2 identities which are in the same resource group with the VM scaleset
          text: az vmss identity remove -g MyResourceGroup -n MyVm --identities readerId writerId
"""

helps['vmss identity show'] = """
    type: command
    short-summary: display VM scaleset's managed identity info.
"""

helps['disk'] = """
    type: group
    short-summary: Manage Azure Managed Disks.
"""

helps['snapshot'] = """
    type: group
    short-summary: Manage point-in-time copies of managed disks, native blobs, or other snapshots.
"""

helps['image'] = """
    type: group
    short-summary: Manage custom virtual machine images.
"""

helps['disk create'] = """
    type: command
    short-summary: Create a managed disk.
    examples:
        - name: Create a managed disk by importing from a blob uri.
          text: >
            az disk create -g MyResourceGroup -n MyDisk --source https://vhd1234.blob.core.windows.net/vhds/osdisk1234.vhd
        - name: Create an empty managed disk.
          text: >
            az disk create -g MyResourceGroup -n MyDisk --size-gb 10
        - name: Create a managed disk by copying an existing disk or snapshot.
          text: >
            az disk create -g MyResourceGroup -n MyDisk2 --source MyDisk
        - name: Create a disk in an availability zone in the region of "East US 2"
          text: >
            az disk create -n MyDisk -g MyResourceGroup --size-gb 10 --location eastus2 --zone 1
"""

helps['disk list'] = """
    type: command
    short-summary: List managed disks.
"""

helps['disk delete'] = """
    type: command
    short-summary: Delete a managed disk.
"""

helps['disk update'] = """
    type: command
    short-summary: Update a managed disk.
"""

helps['disk wait'] = """
    type: command
    short-summary: Place the CLI in a waiting state until a condition of a managed disk is met.
"""


helps['disk grant-access'] = """
    type: command
    short-summary: Grant a resource read access to a managed disk.
"""

helps['disk revoke-access'] = """
    type: command
    short-summary: Revoke a resource's read access to a managed disk.
"""

helps['snapshot create'] = """
    type: command
    short-summary: Create a snapshot.
    examples:
        - name: Create a snapshot by importing from a blob uri.
          text: >
            az snapshot create -g MyResourceGroup -n MySnapshot --source https://vhd1234.blob.core.windows.net/vhds/osdisk1234.vhd
        - name: Create an empty snapshot.
          text: az snapshot create -g MyResourceGroup -n MySnapshot --size-gb 10
        - name: Create a snapshot by copying an existing disk in the same resource group.
          text: az snapshot create -g MyResourceGroup -n MySnapshot2 --source MyDisk
"""

helps['snapshot update'] = """
    type: command
    short-summary: Update a snapshot.
"""

helps['snapshot list'] = """
    type: command
    short-summary: List snapshots.
"""

helps['snapshot grant-access'] = """
    type: command
    short-summary: Grant read access to a snapshot.
"""

helps['snapshot revoke-access'] = """
    type: command
    short-summary: Revoke read access to a snapshot.
"""

helps['image create'] = """
    type: command
    short-summary: Create a custom Virtual Machine Image from managed disks or snapshots.
    examples:
        - name: Create an image from an existing disk.
          text: |
            az image create -g MyResourceGroup -n image1 --os-type Linux \\
                --source /subscriptions/db5eb68e-73e2-4fa8-b18a-0123456789999/resourceGroups/rg1/providers/Microsoft.Compute/snapshots/s1 \\
                --data-snapshot /subscriptions/db5eb68e-73e2-4fa8-b18a-0123456789999/resourceGroups/rg/providers/Microsoft.Compute/snapshots/s2
        - name: Create an image by capturing an existing generalized virtual machine in the same resource group.
          text: az image create -g MyResourceGroup -n image1 --source MyVm1
"""

helps['image list'] = """
    type: command
    short-summary: List custom VM images.
"""

helps['identity'] = """
    type: group
    short-summary: Managed Service Identities
"""

helps['identity list'] = """
    type: command
    short-summary: List Managed Service Identities
"""

helps['identity list-operations'] = """
    type: command
    short-summary: Lists available operations for the Managed Identity provider
"""



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
import uuid
import os

from knack.log import get_logger

from azure.cli.core.commands import LongRunningOperation

from azure.cli.command_modules.vm.custom import set_vm, _compute_client_factory
from azure.cli.command_modules.vm._vm_utils import get_key_vault_base_url, create_keyvault_data_plane_client

_DATA_VOLUME_TYPE = 'DATA'
_STATUS_ENCRYPTED = 'Encrypted'

logger = get_logger(__name__)

vm_extension_info = {
    'Linux': {
        'publisher': os.environ.get('ADE_TEST_EXTENSION_PUBLISHER') or 'Microsoft.Azure.Security',
        'name': os.environ.get('ADE_TEST_EXTENSION_NAME') or 'AzureDiskEncryptionForLinux',
        'version': '0.1'
    },
    'Windows': {
        'publisher': os.environ.get('ADE_TEST_EXTENSION_PUBLISHER') or 'Microsoft.Azure.Security',
        'name': os.environ.get('ADE_TEST_EXTENSION_NAME') or 'AzureDiskEncryption',
        'version': '1.1'
    }
}

vmss_extension_info = {
    'Linux': {
        'publisher': 'Microsoft.Azure.Security',
        'name': vm_extension_info['Linux']['name'],
        'version': '1.1'
    },
    'Windows': {
        'publisher': 'Microsoft.Azure.Security',
        'name': vm_extension_info['Windows']['name'],
        'version': '2.1'
    }
}


def encrypt_vm(cmd, resource_group_name, vm_name,  # pylint: disable=too-many-locals, too-many-statements
               aad_client_id,
               disk_encryption_keyvault,
               aad_client_secret=None, aad_client_cert_thumbprint=None,
               key_encryption_keyvault=None,
               key_encryption_key=None,
               key_encryption_algorithm='RSA-OAEP',
               volume_type=None,
               encrypt_format_all=False):
    from msrestazure.tools import parse_resource_id
    from knack.util import CLIError

    # pylint: disable=no-member
    compute_client = _compute_client_factory(cmd.cli_ctx)
    vm = compute_client.virtual_machines.get(resource_group_name, vm_name)
    os_type = vm.storage_profile.os_disk.os_type.value
    is_linux = _is_linux_vm(os_type)
    extension = vm_extension_info[os_type]
    backup_encryption_settings = vm.storage_profile.os_disk.encryption_settings
    vm_encrypted = backup_encryption_settings.enabled if backup_encryption_settings else False

    # 1. First validate arguments

    if not aad_client_cert_thumbprint and not aad_client_secret:
        raise CLIError('Please provide either --aad-client-cert-thumbprint or --aad-client-secret')

    if volume_type is None:
        if vm.storage_profile.data_disks:
            raise CLIError('VM has data disks, please supply --volume-type')
        else:
            volume_type = 'OS'

    # encryption is not supported on all linux distros, but service never tells you
    # so let us verify at the client side
    if is_linux:
        image_reference = getattr(vm.storage_profile, 'image_reference', None)
        if image_reference:
            result, message = _check_encrypt_is_supported(image_reference, volume_type)
            if not result:
                logger.warning(message)

    # sequence_version should be unique
    sequence_version = uuid.uuid4()

    # retrieve keyvault details
    disk_encryption_keyvault_url = get_key_vault_base_url(
        cmd.cli_ctx, (parse_resource_id(disk_encryption_keyvault))['name'])

    # disk encryption key itself can be further protected, so let us verify
    if key_encryption_key:
        key_encryption_keyvault = key_encryption_keyvault or disk_encryption_keyvault
        if '://' not in key_encryption_key:  # appears a key name
            key_encryption_key = _get_keyvault_key_url(
                cmd.cli_ctx, (parse_resource_id(key_encryption_keyvault))['name'], key_encryption_key)

    # 2. we are ready to provision/update the disk encryption extensions
    # The following logic was mostly ported from xplat-cli
    public_config = {
        'AADClientID': aad_client_id,
        'AADClientCertThumbprint': aad_client_cert_thumbprint,
        'KeyVaultURL': disk_encryption_keyvault_url,
        'VolumeType': volume_type,
        'EncryptionOperation': 'EnableEncryption' if not encrypt_format_all else 'EnableEncryptionFormatAll',
        'KeyEncryptionKeyURL': key_encryption_key,
        'KeyEncryptionAlgorithm': key_encryption_algorithm,
        'SequenceVersion': sequence_version,
    }
    private_config = {
        'AADClientSecret': aad_client_secret if is_linux else (aad_client_secret or '')
    }

    VirtualMachineExtension, DiskEncryptionSettings, KeyVaultSecretReference, KeyVaultKeyReference, SubResource = \
        cmd.get_models('VirtualMachineExtension', 'DiskEncryptionSettings', 'KeyVaultSecretReference',
                       'KeyVaultKeyReference', 'SubResource')

    ext = VirtualMachineExtension(location=vm.location,  # pylint: disable=no-member
                                  publisher=extension['publisher'],
                                  virtual_machine_extension_type=extension['name'],
                                  protected_settings=private_config,
                                  type_handler_version=extension['version'],
                                  settings=public_config,
                                  auto_upgrade_minor_version=True)

    poller = compute_client.virtual_machine_extensions.create_or_update(
        resource_group_name, vm_name, extension['name'], ext)
    poller.result()

    # verify the extension was ok
    extension_result = compute_client.virtual_machine_extensions.get(
        resource_group_name, vm_name, extension['name'], 'instanceView')
    if extension_result.provisioning_state != 'Succeeded':
        raise CLIError('Extension needed for disk encryption was not provisioned correctly')
    if not (extension_result.instance_view.statuses and
            extension_result.instance_view.statuses[0].message):
        raise CLIError('Could not found url pointing to the secret for disk encryption')

    # 3. update VM's storage profile with the secrets
    status_url = extension_result.instance_view.statuses[0].message

    vm = compute_client.virtual_machines.get(resource_group_name, vm_name)
    secret_ref = KeyVaultSecretReference(secret_url=status_url,
                                         source_vault=SubResource(id=disk_encryption_keyvault))

    key_encryption_key_obj = None
    if key_encryption_key:
        key_encryption_key_obj = KeyVaultKeyReference(key_url=key_encryption_key,
                                                      source_vault=SubResource(id=key_encryption_keyvault))

    disk_encryption_settings = DiskEncryptionSettings(disk_encryption_key=secret_ref,
                                                      key_encryption_key=key_encryption_key_obj,
                                                      enabled=True)
    if vm_encrypted:
        # stop the vm before update if the vm is already encrypted
        logger.warning("Deallocating the VM before updating encryption settings...")
        compute_client.virtual_machines.deallocate(resource_group_name, vm_name).result()
        vm = compute_client.virtual_machines.get(resource_group_name, vm_name)

    vm.storage_profile.os_disk.encryption_settings = disk_encryption_settings
    set_vm(cmd, vm)

    if vm_encrypted:
        # and start after the update
        logger.warning("Restarting the VM after the update...")
        compute_client.virtual_machines.start(resource_group_name, vm_name).result()

    if is_linux and volume_type != _DATA_VOLUME_TYPE:
        # TODO: expose a 'wait' command to do the monitor and handle the reboot
        logger.warning("The encryption request was accepted. Please use 'show' command to monitor "
                       "the progress. If you see 'VMRestartPending', please restart the VM, and "
                       "the encryption will finish shortly")


def decrypt_vm(cmd, resource_group_name, vm_name, volume_type=None, force=False):
    from knack.util import CLIError

    compute_client = _compute_client_factory(cmd.cli_ctx)
    vm = compute_client.virtual_machines.get(resource_group_name, vm_name)
    # pylint: disable=no-member
    os_type = vm.storage_profile.os_disk.os_type.value

    # 1. be nice, figure out the default volume type and also verify VM will not be busted
    is_linux = _is_linux_vm(os_type)
    if is_linux:
        if volume_type:
            if not force:
                if volume_type == _DATA_VOLUME_TYPE:
                    status = show_vm_encryption_status(cmd, resource_group_name, vm_name)
                    if status['osDisk'] == _STATUS_ENCRYPTED:
                        raise CLIError("Linux VM's OS disk is encrypted. Disabling encryption on data "
                                       "disk can render the VM unbootable. Use '--force' "
                                       "to ingore the warning")
                else:
                    raise CLIError("Only Data disks can have encryption disabled in a Linux VM. "
                                   "Use '--force' to ingore the warning")
        else:
            volume_type = _DATA_VOLUME_TYPE
    elif volume_type is None:
        if vm.storage_profile.data_disks:
            raise CLIError("VM has data disks, please specify --volume-type")

    extension = vm_extension_info[os_type]
    # sequence_version should be incremented since encryptions occurred before
    sequence_version = uuid.uuid4()

    # 2. update the disk encryption extension
    # The following logic was mostly ported from xplat-cli
    public_config = {
        'VolumeType': volume_type,
        'EncryptionOperation': 'DisableEncryption',
        'SequenceVersion': sequence_version,
    }

    VirtualMachineExtension, DiskEncryptionSettings = cmd.get_models(
        'VirtualMachineExtension', 'DiskEncryptionSettings')

    ext = VirtualMachineExtension(location=vm.location,  # pylint: disable=no-member
                                  publisher=extension['publisher'],
                                  virtual_machine_extension_type=extension['name'],
                                  type_handler_version=extension['version'],
                                  settings=public_config,
                                  auto_upgrade_minor_version=True)

    poller = compute_client.virtual_machine_extensions.create_or_update(resource_group_name,
                                                                        vm_name,
                                                                        extension['name'], ext)
    poller.result()

    # 3. Remove the secret from VM's storage profile
    extension_result = compute_client.virtual_machine_extensions.get(resource_group_name, vm_name,
                                                                     extension['name'],
                                                                     'instanceView')
    if extension_result.provisioning_state != 'Succeeded':
        raise CLIError("Extension updating didn't succeed")

    vm = compute_client.virtual_machines.get(resource_group_name, vm_name)
    disk_encryption_settings = DiskEncryptionSettings(enabled=False)
    vm.storage_profile.os_disk.encryption_settings = disk_encryption_settings
    set_vm(cmd, vm)


def show_vm_encryption_status(cmd, resource_group_name, vm_name):

    encryption_status = {
        'osDisk': 'NotEncrypted',
        'osDiskEncryptionSettings': None,
        'dataDisk': 'NotEncrypted',
        'osType': None
    }
    compute_client = _compute_client_factory(cmd.cli_ctx)
    vm = compute_client.virtual_machines.get(resource_group_name, vm_name)
    # pylint: disable=no-member
    # The following logic was mostly ported from xplat-cli
    os_type = vm.storage_profile.os_disk.os_type.value
    is_linux = _is_linux_vm(os_type)
    encryption_status['osType'] = os_type
    extension = vm_extension_info[os_type]
    extension_result = compute_client.virtual_machine_extensions.get(resource_group_name,
                                                                     vm_name,
                                                                     extension['name'],
                                                                     'instanceView')
    logger.debug(extension_result)
    if extension_result.instance_view.statuses:
        encryption_status['progressMessage'] = extension_result.instance_view.statuses[0].message

    substatus_message = None
    if getattr(extension_result.instance_view, 'substatuses', None):
        substatus_message = extension_result.instance_view.substatuses[0].message

    encryption_status['osDiskEncryptionSettings'] = vm.storage_profile.os_disk.encryption_settings

    import json
    if is_linux:
        try:
            message_object = json.loads(substatus_message)
        except Exception:  # pylint: disable=broad-except
            message_object = None  # might be from outdated extension

        if message_object and ('os' in message_object):
            encryption_status['osDisk'] = message_object['os']
        else:
            encryption_status['osDisk'] = 'Unknown'

        if message_object and 'data' in message_object:
            encryption_status['dataDisk'] = message_object['data']
        else:
            encryption_status['dataDisk'] = 'Unknown'
    else:
        # Windows - get os and data volume encryption state from the vm model
        if (encryption_status['osDiskEncryptionSettings'] and
                encryption_status['osDiskEncryptionSettings'].enabled and
                encryption_status['osDiskEncryptionSettings'].disk_encryption_key and
                encryption_status['osDiskEncryptionSettings'].disk_encryption_key.secret_url):
            encryption_status['osDisk'] = _STATUS_ENCRYPTED
        else:
            encryption_status['osDisk'] = 'Unknown'

        if extension_result.provisioning_state == 'Succeeded':
            volume_type = extension_result.settings.get('VolumeType', None)
            about_data_disk = not volume_type or volume_type.lower() != 'os'
            if about_data_disk and extension_result.settings.get('EncryptionOperation', None) == 'EnableEncryption':
                encryption_status['dataDisk'] = _STATUS_ENCRYPTED

    return encryption_status


def _is_linux_vm(os_type):
    return os_type.lower() == 'linux'


def _get_keyvault_key_url(cli_ctx, keyvault_name, key_name):
    client = create_keyvault_data_plane_client(cli_ctx)
    result = client.get_key(get_key_vault_base_url(cli_ctx, keyvault_name), key_name, '')
    return result.key.kid  # pylint: disable=no-member


def _check_encrypt_is_supported(image_reference, volume_type):
    offer = getattr(image_reference, 'offer', None)
    publisher = getattr(image_reference, 'publisher', None)
    sku = getattr(image_reference, 'sku', None)

    # custom image?
    if not offer or not publisher or not sku:
        return (True, None)

    supported = [
        {
            'offer': 'RHEL',
            'publisher': 'RedHat',
            'sku': '7.2'
        },
        {
            'offer': 'RHEL',
            'publisher': 'RedHat',
            'sku': '7.3'
        },
        {
            'offer': 'CentOS',
            'publisher': 'OpenLogic',
            'sku': '7.2n'
        },
        {
            'offer': 'Ubuntu',
            'publisher': 'Canonical',
            'sku': '14.04'
        },
        {
            'offer': 'Ubuntu',
            'publisher': 'Canonical',
            'sku': '16.04'
        }]

    if volume_type.upper() == _DATA_VOLUME_TYPE:
        supported.append({
            'offer': 'CentOS',
            'publisher': 'OpenLogic',
            'sku': '7.2'
        },)

    for image in supported:
        if (image['publisher'].lower() == publisher.lower() and
                sku.lower().startswith(image['sku'].lower()) and
                offer.lower().startswith(image['offer'].lower())):
            return (True, None)

    sku_list = ['{} {}'.format(a['offer'], a['sku']) for a in supported]

    message = "Encryption might fail as current VM uses a distro not in the known list, which are '{}'".format(sku_list)
    return (False, message)


def _handles_default_volume_type_for_vmss_encryption(is_linux, volume_type, force):
    if is_linux:
        volume_type = volume_type or 'DATA'
        if volume_type != 'DATA':
            msg = 'OS disk encyrption is not yet supported for Linux VM scale sets'
            if force:
                logger.warning(msg)
            else:
                from knack.util import CLIError
                raise CLIError(msg)
    else:
        volume_type = volume_type or 'ALL'
    return volume_type


def encrypt_vmss(cmd, resource_group_name, vmss_name,  # pylint: disable=too-many-locals, too-many-statements
                 disk_encryption_keyvault,
                 key_encryption_keyvault=None,
                 key_encryption_key=None,
                 key_encryption_algorithm='RSA-OAEP',
                 volume_type=None,
                 force=False):
    from msrestazure.tools import parse_resource_id

    # pylint: disable=no-member
    UpgradeMode, VirtualMachineScaleSetExtension, VirtualMachineScaleSetExtensionProfile = cmd.get_models(
        'UpgradeMode', 'VirtualMachineScaleSetExtension', 'VirtualMachineScaleSetExtensionProfile')

    compute_client = _compute_client_factory(cmd.cli_ctx)
    vmss = compute_client.virtual_machine_scale_sets.get(resource_group_name, vmss_name)
    os_type = 'Linux' if vmss.virtual_machine_profile.os_profile.linux_configuration else 'Windows'
    is_linux = _is_linux_vm(os_type)
    extension = vmss_extension_info[os_type]

    # 1. First validate arguments
    volume_type = _handles_default_volume_type_for_vmss_encryption(is_linux, volume_type, force)

    # encryption is not supported on all linux distros, but service never tells you
    # so let us verify at the client side
    if is_linux:
        image_reference = getattr(vmss.virtual_machine_profile.storage_profile, 'image_reference', None)
        if image_reference:
            result, message = _check_encrypt_is_supported(image_reference, volume_type)
            if not result:
                logger.warning(message)

    # retrieve keyvault details
    disk_encryption_keyvault_url = get_key_vault_base_url(cmd.cli_ctx,
                                                          (parse_resource_id(disk_encryption_keyvault))['name'])

    # disk encryption key itself can be further protected, so let us verify
    if key_encryption_key:
        key_encryption_keyvault = key_encryption_keyvault or disk_encryption_keyvault
        if '://' not in key_encryption_key:  # appears a key name
            key_encryption_key = _get_keyvault_key_url(
                cmd.cli_ctx, (parse_resource_id(key_encryption_keyvault))['name'], key_encryption_key)

    #  to avoid bad server errors, ensure the vault has the right configurations
    _verify_keyvault_good_for_encryption(cmd.cli_ctx, disk_encryption_keyvault, key_encryption_keyvault, vmss, force)

    # 2. we are ready to provision/update the disk encryption extensions
    public_config = {
        'KeyVaultURL': disk_encryption_keyvault_url,
        'KeyEncryptionKeyURL': key_encryption_key or '',
        "KeyVaultResourceId": disk_encryption_keyvault,
        "KekVaultResourceId": key_encryption_keyvault if key_encryption_key else '',
        'KeyEncryptionAlgorithm': key_encryption_algorithm if key_encryption_key else '',
        'VolumeType': volume_type,
        'EncryptionOperation': 'EnableEncryption'
    }

    ext = VirtualMachineScaleSetExtension(name=extension['name'],
                                          publisher=extension['publisher'],
                                          type=extension['name'],
                                          type_handler_version=extension['version'],
                                          settings=public_config,
                                          auto_upgrade_minor_version=True,
                                          force_update_tag=uuid.uuid4())
    if not vmss.virtual_machine_profile.extension_profile:
        vmss.virtual_machine_profile.extension_profile = VirtualMachineScaleSetExtensionProfile(extensions=[])
    vmss.virtual_machine_profile.extension_profile.extensions.append(ext)
    poller = compute_client.virtual_machine_scale_sets.create_or_update(resource_group_name, vmss_name, vmss)
    LongRunningOperation(cmd.cli_ctx)(poller)
    _show_post_action_message(resource_group_name, vmss.name, vmss.upgrade_policy.mode == UpgradeMode.manual, True)


def decrypt_vmss(cmd, resource_group_name, vmss_name, volume_type=None, force=False):
    UpgradeMode, VirtualMachineScaleSetExtension = cmd.get_models('UpgradeMode', 'VirtualMachineScaleSetExtension')
    compute_client = _compute_client_factory(cmd.cli_ctx)
    vmss = compute_client.virtual_machine_scale_sets.get(resource_group_name, vmss_name)
    os_type = 'Linux' if vmss.virtual_machine_profile.os_profile.linux_configuration else 'Windows'
    is_linux = _is_linux_vm(os_type)
    extension = vmss_extension_info[os_type]

    # 1. be nice, figure out the default volume type
    volume_type = _handles_default_volume_type_for_vmss_encryption(is_linux, volume_type, force)

    # 2. update the disk encryption extension
    public_config = {
        'VolumeType': volume_type,
        'EncryptionOperation': 'DisableEncryption',
    }

    ext = VirtualMachineScaleSetExtension(name=extension['name'],
                                          publisher=extension['publisher'],
                                          type=extension['name'],
                                          type_handler_version=extension['version'],
                                          settings=public_config,
                                          auto_upgrade_minor_version=True,
                                          force_update_tag=uuid.uuid4())
    if (not vmss.virtual_machine_profile.extension_profile or
            not vmss.virtual_machine_profile.extension_profile.extensions):
        extensions = []
    else:
        extensions = vmss.virtual_machine_profile.extension_profile.extensions

    ade_extension = [x for x in extensions if
                     x.type.lower() == extension['name'].lower() and x.publisher.lower() == extension['publisher'].lower()]  # pylint: disable=line-too-long
    if not ade_extension:
        from knack.util import CLIError
        raise CLIError("VM scale set '{}' was not encrypted".format(vmss_name))

    index = vmss.virtual_machine_profile.extension_profile.extensions.index(ade_extension[0])
    vmss.virtual_machine_profile.extension_profile.extensions[index] = ext
    poller = compute_client.virtual_machine_scale_sets.create_or_update(resource_group_name, vmss_name, vmss)
    LongRunningOperation(cmd.cli_ctx)(poller)
    _show_post_action_message(resource_group_name, vmss.name, vmss.upgrade_policy.mode == UpgradeMode.manual, False)


def _show_post_action_message(resource_group_name, vmss_name, maunal_mode, enable):
    msg = ''
    if maunal_mode:
        msg = ("With manual upgrade mode, you will need to run 'az vmss update-instances -g {} -n {} "
               "--instance-ids \"*\"' to propagate the change.\n".format(resource_group_name, vmss_name))
    msg += ("Note, {} encryption will take a while to finish. Please query the status using "
            "'az vmss encryption show -g {} -n {}'. For Linux VM, you will lose the access during the period".format(
                'enabling' if enable else 'disabling', resource_group_name, vmss_name))
    logger.warning(msg)


def show_vmss_encryption_status(cmd, resource_group_name, vmss_name):
    client = _compute_client_factory(cmd.cli_ctx)
    vm_instances = list(client.virtual_machine_scale_set_vms.list(resource_group_name, vmss_name,
                                                                  select='instanceView', expand='instanceView'))
    result = []
    for instance in vm_instances:
        view = instance.instance_view
        disk_infos = []
        vm_enc_info = {
            'id': instance.id,
            'disks': disk_infos
        }
        for div in view.disks:
            disk_infos.append({
                'name': div.name,
                'encryptionSettings': div.encryption_settings,
                'statuses': [x for x in (div.statuses or []) if (x.code or '').startswith('EncryptionState')]
            })

        result.append(vm_enc_info)
    return result


def _verify_keyvault_good_for_encryption(cli_ctx, disk_vault_id, kek_vault_id, vmss, force):
    def _report_client_side_validation_error(msg):
        if force:
            logger.warning(msg)
        else:
            from knack.util import CLIError
            raise CLIError(msg)

    from azure.cli.core.commands.client_factory import get_mgmt_service_client
    from azure.mgmt.keyvault import KeyVaultManagementClient
    from msrestazure.tools import parse_resource_id

    client = get_mgmt_service_client(cli_ctx, KeyVaultManagementClient).vaults
    disk_vault_resource_info = parse_resource_id(disk_vault_id)
    key_vault = client.get(disk_vault_resource_info['resource_group'], disk_vault_resource_info['name'])

    # ensure vault has 'EnabledForDiskEncryption' permission
    if not key_vault.properties.enabled_for_disk_encryption:
        _report_client_side_validation_error("keyvault '{}' is not enabled for disk encryption. ".format(
            disk_vault_resource_info['resource_name']))

    if kek_vault_id:
        kek_vault_info = parse_resource_id(kek_vault_id)
        if disk_vault_resource_info['name'].lower() != kek_vault_info['name'].lower():
            client.get(kek_vault_info['resource_group'], kek_vault_info['name'])

    # verify subscription mataches
    vmss_resource_info = parse_resource_id(vmss.id)
    if vmss_resource_info['subscription'].lower() != disk_vault_resource_info['subscription'].lower():
        _report_client_side_validation_error(
            "VM scale set's subscription doesn't match keyvault's subscription. Encryption might fail")

    # verify region matches
    if key_vault.location.replace(' ', '').lower() != vmss.location.replace(' ', '').lower():
        _report_client_side_validation_error(
            "VM scale set's region doesn't match keyvault's region. Encryption might fail")



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

# flake8: noqa
# noqa
# pylint: skip-file


def get_default_diag_config(is_windows):
    if is_windows:
        return {
            "WadCfg": {
                "DiagnosticMonitorConfiguration": {
                    "overallQuotaInMB": "4096",
                    "DiagnosticInfrastructureLogs": {
                        "scheduledTransferLogLevelFilter": "Error"
                    },
                    "WindowsEventLog": {
                        "scheduledTransferPeriod": "PT1M",
                        "DataSource": [
                            {"name": "Application!*[System[(Level = 1) or (Level = 2)]]"},
                            {"name": "Security!*[System[(Level = 1 or Level = 2)]]"},
                            {"name": "System!*[System[(Level = 1 or Level = 2)]]"}
                        ]
                    },
                    "PerformanceCounters": {
                        "scheduledTransferPeriod": "PT1M",
                        "PerformanceCounterConfiguration": [
                            {
                                "counterSpecifier": "\\Processor(_Total)\\% Processor Time",
                                "sampleRate": "PT15S",
                                "unit": "Percent",
                                "annotation": [
                                    {
                                        "displayName": "CPU utilization",
                                        "locale": "en-us"
                                    }
                                ]
                            },
                            {
                                "counterSpecifier": "\\Processor(_Total)\\% Privileged Time",
                                "sampleRate": "PT15S",
                                "unit": "Percent",
                                "annotation": [
                                    {
                                        "displayName": "CPU privileged time",
                                        "locale": "en-us"
                                    }
                                ]
                            },
                            {
                                "counterSpecifier": "\\Processor(_Total)\\% User Time",
                                "sampleRate": "PT15S",
                                "unit": "Percent",
                                "annotation": [
                                    {
                                        "displayName": "CPU user time",
                                        "locale": "en-us"
                                    }
                                ]
                            },
                            {
                                "counterSpecifier": "\\Processor Information(_Total)\\Processor Frequency",
                                "sampleRate": "PT15S",
                                "unit": "Count",
                                "annotation": [
                                    {
                                        "displayName": "CPU frequency",
                                        "locale": "en-us"
                                    }
                                ]
                            },
                            {
                                "counterSpecifier": "\\System\\Processes",
                                "sampleRate": "PT15S",
                                "unit": "Count",
                                "annotation": [
                                    {
                                        "displayName": "Processes",
                                        "locale": "en-us"
                                    }
                                ]
                            },
                            {
                                "counterSpecifier": "\\Process(_Total)\\Thread Count",
                                "sampleRate": "PT15S",
                                "unit": "Count",
                                "annotation": [
                                    {
                                        "displayName": "Threads",
                                        "locale": "en-us"
                                    }
                                ]
                            },
                            {
                                "counterSpecifier": "\\Process(_Total)\\Handle Count",
                                "sampleRate": "PT15S",
                                "unit": "Count",
                                "annotation": [
                                    {
                                        "displayName": "Handles",
                                        "locale": "en-us"
                                    }
                                ]
                            },
                            {
                                "counterSpecifier": "\\Memory\\% Committed Bytes In Use",
                                "sampleRate": "PT15S",
                                "unit": "Percent",
                                "annotation": [
                                    {
                                        "displayName": "Memory usage",
                                        "locale": "en-us"
                                    }
                                ]
                            },
                            {
                                "counterSpecifier": "\\Memory\\Available Bytes",
                                "sampleRate": "PT15S",
                                "unit": "Bytes",
                                "annotation": [
                                    {
                                        "displayName": "Memory available",
                                        "locale": "en-us"
                                    }
                                ]
                            },
                            {
                                "counterSpecifier": "\\Memory\\Committed Bytes",
                                "sampleRate": "PT15S",
                                "unit": "Bytes",
                                "annotation": [
                                    {
                                        "displayName": "Memory committed",
                                        "locale": "en-us"
                                    }
                                ]
                            },
                            {
                                "counterSpecifier": "\\Memory\\Commit Limit",
                                "sampleRate": "PT15S",
                                "unit": "Bytes",
                                "annotation": [
                                    {
                                        "displayName": "Memory commit limit",
                                        "locale": "en-us"
                                    }
                                ]
                            },
                            {
                                "counterSpecifier": "\\PhysicalDisk(_Total)\\% Disk Time",
                                "sampleRate": "PT15S",
                                "unit": "Percent",
                                "annotation": [
                                    {
                                        "displayName": "Disk active time",
                                        "locale": "en-us"
                                    }
                                ]
                            },
                            {
                                "counterSpecifier": "\\PhysicalDisk(_Total)\\% Disk Read Time",
                                "sampleRate": "PT15S",
                                "unit": "Percent",
                                "annotation": [
                                    {
                                        "displayName": "Disk active read time",
                                        "locale": "en-us"
                                    }
                                ]
                            },
                            {
                                "counterSpecifier": "\\PhysicalDisk(_Total)\\% Disk Write Time",
                                "sampleRate": "PT15S",
                                "unit": "Percent",
                                "annotation": [
                                    {
                                        "displayName": "Disk active write time",
                                        "locale": "en-us"
                                    }
                                ]
                            },
                            {
                                "counterSpecifier": "\\PhysicalDisk(_Total)\\Disk Transfers/sec",
                                "sampleRate": "PT15S",
                                "unit": "CountPerSecond",
                                "annotation": [
                                    {
                                        "displayName": "Disk operations",
                                        "locale": "en-us"
                                    }
                                ]
                            },
                            {
                                "counterSpecifier": "\\PhysicalDisk(_Total)\\Disk Reads/sec",
                                "sampleRate": "PT15S",
                                "unit": "CountPerSecond",
                                "annotation": [
                                    {
                                        "displayName": "Disk read operations",
                                        "locale": "en-us"
                                    }
                                ]
                            },
                            {
                                "counterSpecifier": "\\PhysicalDisk(_Total)\\Disk Writes/sec",
                                "sampleRate": "PT15S",
                                "unit": "CountPerSecond",
                                "annotation": [
                                    {
                                        "displayName": "Disk write operations",
                                        "locale": "en-us"
                                    }
                                ]
                            },
                            {
                                "counterSpecifier": "\\PhysicalDisk(_Total)\\Disk Bytes/sec",
                                "sampleRate": "PT15S",
                                "unit": "BytesPerSecond",
                                "annotation": [
                                    {
                                        "displayName": "Disk speed",
                                        "locale": "en-us"
                                    }
                                ]
                            },
                            {
                                "counterSpecifier": "\\PhysicalDisk(_Total)\\Disk Read Bytes/sec",
                                "sampleRate": "PT15S",
                                "unit": "BytesPerSecond",
                                "annotation": [
                                    {
                                        "displayName": "Disk read speed",
                                        "locale": "en-us"
                                    }
                                ]
                            },
                            {
                                "counterSpecifier": "\\PhysicalDisk(_Total)\\Disk Write Bytes/sec",
                                "sampleRate": "PT15S",
                                "unit": "BytesPerSecond",
                                "annotation": [
                                    {
                                        "displayName": "Disk write speed",
                                        "locale": "en-us"
                                    }
                                ]
                            },
                            {
                                "counterSpecifier": "\\LogicalDisk(_Total)\\% Free Space",
                                "sampleRate": "PT15S",
                                "unit": "Percent",
                                "annotation": [
                                    {
                                        "displayName": "Disk free space (percentage)",
                                        "locale": "en-us"
                                    }
                                ]
                            }
                        ]
                    },
                    "Metrics": {
                        "resourceId": "[variables('wadmetricsresourceid')]",
                        "MetricAggregation": [
                            {
                                "scheduledTransferPeriod": "PT1H"
                            },
                            {
                                "scheduledTransferPeriod": "PT1M"
                            }
                        ]
                    }
                }
            },
            "storageAccount": "__DIAGNOSTIC_STORAGE_ACCOUNT__"
        }
    else:
        return {
            "StorageAccount": "__DIAGNOSTIC_STORAGE_ACCOUNT__",
            "ladCfg": {
                "diagnosticMonitorConfiguration": {
                    "eventVolume": "Medium", 
                    "metrics": {
                        "metricAggregation": [
                            {
                                "scheduledTransferPeriod": "PT1H"
                            }, 
                            {
                                "scheduledTransferPeriod": "PT1M"
                            }
                        ], 
                        "resourceId": "__VM_OR_VMSS_RESOURCE_ID__"
                    }, 
                    "performanceCounters": {
                        "performanceCounterConfiguration": [
                            {
                                "annotation": [
                                    {
                                        "displayName": "Disk read guest OS", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "disk", 
                                "condition": "IsAggregate=TRUE", 
                                "counter": "readbytespersecond", 
                                "counterSpecifier": "/builtin/disk/readbytespersecond", 
                                "type": "builtin", 
                                "unit": "BytesPerSecond"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "Disk writes", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "disk", 
                                "condition": "IsAggregate=TRUE", 
                                "counter": "writespersecond", 
                                "counterSpecifier": "/builtin/disk/writespersecond", 
                                "type": "builtin", 
                                "unit": "CountPerSecond"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "Disk transfer time", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "disk", 
                                "condition": "IsAggregate=TRUE", 
                                "counter": "averagetransfertime", 
                                "counterSpecifier": "/builtin/disk/averagetransfertime", 
                                "type": "builtin", 
                                "unit": "Seconds"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "Disk transfers", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "disk", 
                                "condition": "IsAggregate=TRUE", 
                                "counter": "transferspersecond", 
                                "counterSpecifier": "/builtin/disk/transferspersecond", 
                                "type": "builtin", 
                                "unit": "CountPerSecond"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "Disk write guest OS", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "disk", 
                                "condition": "IsAggregate=TRUE", 
                                "counter": "writebytespersecond", 
                                "counterSpecifier": "/builtin/disk/writebytespersecond", 
                                "type": "builtin", 
                                "unit": "BytesPerSecond"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "Disk read time", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "disk", 
                                "condition": "IsAggregate=TRUE", 
                                "counter": "averagereadtime", 
                                "counterSpecifier": "/builtin/disk/averagereadtime", 
                                "type": "builtin", 
                                "unit": "Seconds"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "Disk write time", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "disk", 
                                "condition": "IsAggregate=TRUE", 
                                "counter": "averagewritetime", 
                                "counterSpecifier": "/builtin/disk/averagewritetime", 
                                "type": "builtin", 
                                "unit": "Seconds"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "Disk total bytes", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "disk", 
                                "condition": "IsAggregate=TRUE", 
                                "counter": "bytespersecond", 
                                "counterSpecifier": "/builtin/disk/bytespersecond", 
                                "type": "builtin", 
                                "unit": "BytesPerSecond"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "Disk reads", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "disk", 
                                "condition": "IsAggregate=TRUE", 
                                "counter": "readspersecond", 
                                "counterSpecifier": "/builtin/disk/readspersecond", 
                                "type": "builtin", 
                                "unit": "CountPerSecond"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "Disk queue length", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "disk", 
                                "condition": "IsAggregate=TRUE", 
                                "counter": "averagediskqueuelength", 
                                "counterSpecifier": "/builtin/disk/averagediskqueuelength", 
                                "type": "builtin", 
                                "unit": "Count"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "Network in guest OS", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "network", 
                                "counter": "bytesreceived", 
                                "counterSpecifier": "/builtin/network/bytesreceived", 
                                "type": "builtin", 
                                "unit": "Bytes"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "Network total bytes", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "network", 
                                "counter": "bytestotal", 
                                "counterSpecifier": "/builtin/network/bytestotal", 
                                "type": "builtin", 
                                "unit": "Bytes"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "Network out guest OS", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "network", 
                                "counter": "bytestransmitted", 
                                "counterSpecifier": "/builtin/network/bytestransmitted", 
                                "type": "builtin", 
                                "unit": "Bytes"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "Network collisions", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "network", 
                                "counter": "totalcollisions", 
                                "counterSpecifier": "/builtin/network/totalcollisions", 
                                "type": "builtin", 
                                "unit": "Count"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "Packets received errors", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "network", 
                                "counter": "totalrxerrors", 
                                "counterSpecifier": "/builtin/network/totalrxerrors", 
                                "type": "builtin", 
                                "unit": "Count"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "Packets sent", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "network", 
                                "counter": "packetstransmitted", 
                                "counterSpecifier": "/builtin/network/packetstransmitted", 
                                "type": "builtin", 
                                "unit": "Count"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "Packets received", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "network", 
                                "counter": "packetsreceived", 
                                "counterSpecifier": "/builtin/network/packetsreceived", 
                                "type": "builtin", 
                                "unit": "Count"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "Packets sent errors", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "network", 
                                "counter": "totaltxerrors", 
                                "counterSpecifier": "/builtin/network/totaltxerrors", 
                                "type": "builtin", 
                                "unit": "Count"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "Filesystem transfers/sec", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "filesystem", 
                                "condition": "IsAggregate=TRUE", 
                                "counter": "transferspersecond", 
                                "counterSpecifier": "/builtin/filesystem/transferspersecond", 
                                "type": "builtin", 
                                "unit": "CountPerSecond"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "Filesystem % free space", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "filesystem", 
                                "condition": "IsAggregate=TRUE", 
                                "counter": "percentfreespace", 
                                "counterSpecifier": "/builtin/filesystem/percentfreespace", 
                                "type": "builtin", 
                                "unit": "Percent"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "Filesystem % used space", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "filesystem", 
                                "condition": "IsAggregate=TRUE", 
                                "counter": "percentusedspace", 
                                "counterSpecifier": "/builtin/filesystem/percentusedspace", 
                                "type": "builtin", 
                                "unit": "Percent"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "Filesystem used space", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "filesystem", 
                                "condition": "IsAggregate=TRUE", 
                                "counter": "usedspace", 
                                "counterSpecifier": "/builtin/filesystem/usedspace", 
                                "type": "builtin", 
                                "unit": "Bytes"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "Filesystem read bytes/sec", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "filesystem", 
                                "condition": "IsAggregate=TRUE", 
                                "counter": "bytesreadpersecond", 
                                "counterSpecifier": "/builtin/filesystem/bytesreadpersecond", 
                                "type": "builtin", 
                                "unit": "CountPerSecond"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "Filesystem free space", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "filesystem", 
                                "condition": "IsAggregate=TRUE", 
                                "counter": "freespace", 
                                "counterSpecifier": "/builtin/filesystem/freespace", 
                                "type": "builtin", 
                                "unit": "Bytes"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "Filesystem % free inodes", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "filesystem", 
                                "condition": "IsAggregate=TRUE", 
                                "counter": "percentfreeinodes", 
                                "counterSpecifier": "/builtin/filesystem/percentfreeinodes", 
                                "type": "builtin", 
                                "unit": "Percent"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "Filesystem bytes/sec", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "filesystem", 
                                "condition": "IsAggregate=TRUE", 
                                "counter": "bytespersecond", 
                                "counterSpecifier": "/builtin/filesystem/bytespersecond", 
                                "type": "builtin", 
                                "unit": "BytesPerSecond"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "Filesystem reads/sec", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "filesystem", 
                                "condition": "IsAggregate=TRUE", 
                                "counter": "readspersecond", 
                                "counterSpecifier": "/builtin/filesystem/readspersecond", 
                                "type": "builtin", 
                                "unit": "CountPerSecond"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "Filesystem write bytes/sec", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "filesystem", 
                                "condition": "IsAggregate=TRUE", 
                                "counter": "byteswrittenpersecond", 
                                "counterSpecifier": "/builtin/filesystem/byteswrittenpersecond", 
                                "type": "builtin", 
                                "unit": "CountPerSecond"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "Filesystem writes/sec", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "filesystem", 
                                "condition": "IsAggregate=TRUE", 
                                "counter": "writespersecond", 
                                "counterSpecifier": "/builtin/filesystem/writespersecond", 
                                "type": "builtin", 
                                "unit": "CountPerSecond"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "Filesystem % used inodes", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "filesystem", 
                                "condition": "IsAggregate=TRUE", 
                                "counter": "percentusedinodes", 
                                "counterSpecifier": "/builtin/filesystem/percentusedinodes", 
                                "type": "builtin", 
                                "unit": "Percent"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "CPU IO wait time", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "processor", 
                                "condition": "IsAggregate=TRUE", 
                                "counter": "percentiowaittime", 
                                "counterSpecifier": "/builtin/processor/percentiowaittime", 
                                "type": "builtin", 
                                "unit": "Percent"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "CPU user time", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "processor", 
                                "condition": "IsAggregate=TRUE", 
                                "counter": "percentusertime", 
                                "counterSpecifier": "/builtin/processor/percentusertime", 
                                "type": "builtin", 
                                "unit": "Percent"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "CPU nice time", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "processor", 
                                "condition": "IsAggregate=TRUE", 
                                "counter": "percentnicetime", 
                                "counterSpecifier": "/builtin/processor/percentnicetime", 
                                "type": "builtin", 
                                "unit": "Percent"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "CPU percentage guest OS", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "processor", 
                                "condition": "IsAggregate=TRUE", 
                                "counter": "percentprocessortime", 
                                "counterSpecifier": "/builtin/processor/percentprocessortime", 
                                "type": "builtin", 
                                "unit": "Percent"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "CPU interrupt time", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "processor", 
                                "condition": "IsAggregate=TRUE", 
                                "counter": "percentinterrupttime", 
                                "counterSpecifier": "/builtin/processor/percentinterrupttime", 
                                "type": "builtin", 
                                "unit": "Percent"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "CPU idle time", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "processor", 
                                "condition": "IsAggregate=TRUE", 
                                "counter": "percentidletime", 
                                "counterSpecifier": "/builtin/processor/percentidletime", 
                                "type": "builtin", 
                                "unit": "Percent"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "CPU privileged time", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "processor", 
                                "condition": "IsAggregate=TRUE", 
                                "counter": "percentprivilegedtime", 
                                "counterSpecifier": "/builtin/processor/percentprivilegedtime", 
                                "type": "builtin", 
                                "unit": "Percent"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "Memory available", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "memory", 
                                "counter": "availablememory", 
                                "counterSpecifier": "/builtin/memory/availablememory", 
                                "type": "builtin", 
                                "unit": "Bytes"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "Swap percent used", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "memory", 
                                "counter": "percentusedswap", 
                                "counterSpecifier": "/builtin/memory/percentusedswap", 
                                "type": "builtin", 
                                "unit": "Percent"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "Memory used", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "memory", 
                                "counter": "usedmemory", 
                                "counterSpecifier": "/builtin/memory/usedmemory", 
                                "type": "builtin", 
                                "unit": "Bytes"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "Page reads", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "memory", 
                                "counter": "pagesreadpersec", 
                                "counterSpecifier": "/builtin/memory/pagesreadpersec", 
                                "type": "builtin", 
                                "unit": "CountPerSecond"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "Swap available", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "memory", 
                                "counter": "availableswap", 
                                "counterSpecifier": "/builtin/memory/availableswap", 
                                "type": "builtin", 
                                "unit": "Bytes"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "Swap percent available", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "memory", 
                                "counter": "percentavailableswap", 
                                "counterSpecifier": "/builtin/memory/percentavailableswap", 
                                "type": "builtin", 
                                "unit": "Percent"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "Mem. percent available", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "memory", 
                                "counter": "percentavailablememory", 
                                "counterSpecifier": "/builtin/memory/percentavailablememory", 
                                "type": "builtin", 
                                "unit": "Percent"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "Pages", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "memory", 
                                "counter": "pagespersec", 
                                "counterSpecifier": "/builtin/memory/pagespersec", 
                                "type": "builtin", 
                                "unit": "CountPerSecond"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "Swap used", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "memory", 
                                "counter": "usedswap", 
                                "counterSpecifier": "/builtin/memory/usedswap", 
                                "type": "builtin", 
                                "unit": "Bytes"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "Memory percentage", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "memory", 
                                "counter": "percentusedmemory", 
                                "counterSpecifier": "/builtin/memory/percentusedmemory", 
                                "type": "builtin", 
                                "unit": "Percent"
                            }, 
                            {
                                "annotation": [
                                    {
                                        "displayName": "Page writes", 
                                        "locale": "en-us"
                                    }
                                ], 
                                "class": "memory", 
                                "counter": "pageswrittenpersec", 
                                "counterSpecifier": "/builtin/memory/pageswrittenpersec", 
                                "type": "builtin", 
                                "unit": "CountPerSecond"
                            }
                        ]
                    }, 
                    "syslogEvents": {
                        "syslogEventConfiguration": {
                            "LOG_AUTH": "LOG_DEBUG", 
                            "LOG_AUTHPRIV": "LOG_DEBUG", 
                            "LOG_CRON": "LOG_DEBUG", 
                            "LOG_DAEMON": "LOG_DEBUG", 
                            "LOG_FTP": "LOG_DEBUG", 
                            "LOG_KERN": "LOG_DEBUG", 
                            "LOG_LOCAL0": "LOG_DEBUG", 
                            "LOG_LOCAL1": "LOG_DEBUG", 
                            "LOG_LOCAL2": "LOG_DEBUG", 
                            "LOG_LOCAL3": "LOG_DEBUG", 
                            "LOG_LOCAL4": "LOG_DEBUG", 
                            "LOG_LOCAL5": "LOG_DEBUG", 
                            "LOG_LOCAL6": "LOG_DEBUG", 
                            "LOG_LOCAL7": "LOG_DEBUG", 
                            "LOG_LPR": "LOG_DEBUG", 
                            "LOG_MAIL": "LOG_DEBUG", 
                            "LOG_NEWS": "LOG_DEBUG", 
                            "LOG_SYSLOG": "LOG_DEBUG", 
                            "LOG_USER": "LOG_DEBUG", 
                            "LOG_UUCP": "LOG_DEBUG"
                        }
                    }
                }, 
                "sampleRateInSeconds": 15
            }
        }



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from azure.cli.command_modules.vm._client_factory import (cf_vm, cf_avail_set, cf_ni,
                                                          cf_vm_ext,
                                                          cf_vm_ext_image, cf_vm_image, cf_usage,
                                                          cf_vmss, cf_vmss_vm,
                                                          cf_vm_sizes, cf_disks, cf_snapshots,
                                                          cf_images, cf_run_commands,
                                                          cf_rolling_upgrade_commands,
                                                          cf_msi_user_identities_operations,
                                                          cf_msi_operations_operations)
from azure.cli.command_modules.vm._format import (
    transform_ip_addresses, transform_vm, transform_vm_create_output, transform_vm_usage_list, transform_vm_list,
    transform_sku_for_table_output, transform_disk_show_table_output, transform_extension_show_table_output,
    get_vmss_table_output_transformer)
from azure.cli.command_modules.vm._validators import (
    process_vm_create_namespace, process_vmss_create_namespace, process_image_create_namespace,
    process_disk_or_snapshot_create_namespace, process_disk_encryption_namespace, process_assign_identity_namespace,
    process_msi_namespace, process_remove_identity_namespace, process_vm_secret_format)

from azure.cli.core.commands import DeploymentOutputLongRunningOperation, CliCommandType
from azure.cli.core.commands.arm import deployment_validate_table_format
from azure.cli.core.util import empty_on_404


# pylint: disable=line-too-long, too-many-statements
def load_command_table(self, _):

    custom_tmpl = 'azure.cli.command_modules.vm.custom#{}'

    compute_custom = CliCommandType(operations_tmpl=custom_tmpl)

    compute_disk_encryption_custom = CliCommandType(
        operations_tmpl='azure.cli.command_modules.vm.disk_encryption#{}',
        operation_group='virtual_machines'
    )

    compute_availset_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.compute.operations.availability_sets_operations#AvailabilitySetsOperations.{}',
        client_factory=cf_avail_set,
        operation_group='availability_sets'
    )

    compute_disk_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.compute.operations.disks_operations#DisksOperations.{}',
        client_factory=cf_disks,
        operation_group='disks'
    )

    compute_identity_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.msi.operations.user_assigned_identities_operations#UserAssignedIdentitiesOperations.{}',
        client_factory=cf_msi_user_identities_operations
    )

    compute_image_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.compute.operations.images_operations#ImagesOperations.{}',
        client_factory=cf_images
    )

    compute_snapshot_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.compute.operations.snapshots_operations#SnapshotsOperations.{}',
        client_factory=cf_snapshots
    )

    compute_vm_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.compute.operations.virtual_machines_operations#VirtualMachinesOperations.{}',
        client_factory=cf_vm
    )

    compute_vm_extension_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.compute.operations.virtual_machine_extensions_operations#VirtualMachineExtensionsOperations.{}',
        client_factory=cf_vm_ext
    )

    compute_vm_extension_image_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.compute.operations.virtual_machine_extension_images_operations#VirtualMachineExtensionImagesOperations.{}',
        client_factory=cf_vm_ext_image
    )

    compute_vm_image_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.compute.operations.virtual_machine_images_operations#VirtualMachineImagesOperations.{}',
        client_factory=cf_vm_image
    )

    compute_vm_usage_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.compute.operations.usage_operations#UsageOperations.{}',
        client_factory=cf_usage
    )

    compute_vm_run_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.compute.operations.virtual_machine_run_commands_operations#VirtualMachineRunCommandsOperations.{}',
        client_factory=cf_run_commands,
        min_api='2017-03-30',
        max_api='2017-03-30'
    )

    compute_vm_size_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.compute.operations.virtual_machine_sizes_operations#VirtualMachineSizesOperations.{}',
        client_factory=cf_vm_sizes
    )

    compute_vmss_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.compute.operations.virtual_machine_scale_sets_operations#VirtualMachineScaleSetsOperations.{}',
        client_factory=cf_vmss,
        operation_group='virtual_machine_scale_sets'
    )

    compute_vmss_rolling_upgrade_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.compute.operations.virtual_machine_scale_set_rolling_upgrades_operations#VirtualMachineScaleSetRollingUpgradesOperations.{}',
        client_factory=cf_rolling_upgrade_commands,
        operation_group='virtual_machine_scale_sets'
    )

    compute_vmss_vm_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.compute.operations.virtual_machine_scale_set_vms_operations#VirtualMachineScaleSetVMsOperations.{}',
        client_factory=cf_vmss_vm,
        operation_group='virtual_machine_scale_sets'
    )

    network_nic_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.network.operations.network_interfaces_operations#NetworkInterfacesOperations.{}',
        client_factory=cf_ni
    )

    with self.command_group('disk', compute_disk_sdk, operation_group='disks', min_api='2017-03-30') as g:
        g.custom_command('create', 'create_managed_disk', supports_no_wait=True, table_transformer=transform_disk_show_table_output, validator=process_disk_or_snapshot_create_namespace)
        g.command('delete', 'delete', supports_no_wait=True, confirmation=True)
        g.custom_command('grant-access', 'grant_disk_access')
        g.custom_command('list', 'list_managed_disks', table_transformer='[].' + transform_disk_show_table_output)
        g.command('revoke-access', 'revoke_access')
        g.command('show', 'get', exception_handler=empty_on_404, table_transformer=transform_disk_show_table_output)
        g.generic_update_command('update', custom_func_name='update_managed_disk', setter_arg_name='disk', supports_no_wait=True)
        g.generic_wait_command('wait')

    # TODO move to its own command module https://github.com/Azure/azure-cli/issues/5105
    with self.command_group('identity', compute_identity_sdk, min_api='2017-12-01') as g:
        g.command('create', 'create_or_update', validator=process_msi_namespace)
        g.command('show', 'get')
        g.command('delete', 'delete')
        g.custom_command('list', 'list_user_assigned_identities')
        g.command('list-operations', 'list', operations_tmpl='azure.mgmt.msi.operations.operations#Operations.{}', client_factory=cf_msi_operations_operations)

    with self.command_group('image', compute_image_sdk, min_api='2016-04-30-preview') as g:
        g.custom_command('create', 'create_image', validator=process_image_create_namespace)
        g.custom_command('list', 'list_images')
        g.command('show', 'get', exception_handler=empty_on_404)
        g.command('delete', 'delete')

    with self.command_group('snapshot', compute_snapshot_sdk, operation_group='snapshots', min_api='2016-04-30-preview') as g:
        g.custom_command('create', 'create_snapshot', validator=process_disk_or_snapshot_create_namespace)
        g.command('delete', 'delete')
        g.custom_command('grant-access', 'grant_snapshot_access')
        g.custom_command('list', 'list_snapshots')
        g.command('revoke-access', 'revoke_access')
        g.command('show', 'get', exception_handler=empty_on_404)
        g.generic_update_command('update', custom_func_name='update_snapshot', setter_arg_name='snapshot')

    with self.command_group('vm', compute_vm_sdk) as g:
        g.custom_command('identity assign', 'assign_vm_identity', validator=process_assign_identity_namespace)
        g.custom_command('identity remove', 'remove_vm_identity', validator=process_remove_identity_namespace, min_api='2017-12-01')
        g.custom_command('identity show', 'show_vm_identity')

        g.custom_command('capture', 'capture_vm')
        g.custom_command('create', 'create_vm', transform=transform_vm_create_output, supports_no_wait=True, table_transformer=deployment_validate_table_format, validator=process_vm_create_namespace)
        g.command('convert', 'convert_to_managed_disks', min_api='2016-04-30-preview')
        g.command('deallocate', 'deallocate', supports_no_wait=True)
        g.command('delete', 'delete', confirmation=True, supports_no_wait=True)
        g.command('generalize', 'generalize', supports_no_wait=True)
        g.custom_command('get-instance-view', 'get_instance_view', table_transformer='{Name:name, ResourceGroup:resourceGroup, Location:location, ProvisioningState:provisioningState, PowerState:instanceView.statuses[1].displayStatus}')
        g.custom_command('list', 'list_vm', table_transformer=transform_vm_list)
        g.custom_command('list-ip-addresses', 'list_vm_ip_addresses', table_transformer=transform_ip_addresses)
        g.command('list-sizes', 'list', command_type=compute_vm_size_sdk)
        g.custom_command('list-skus', 'list_skus', table_transformer=transform_sku_for_table_output, min_api='2017-03-30')
        g.command('list-usage', 'list', command_type=compute_vm_usage_sdk, transform=transform_vm_usage_list, table_transformer='[].{Name:localName, CurrentValue:currentValue, Limit:limit}')
        g.command('list-vm-resize-options', 'list_available_sizes')
        g.custom_command('open-port', 'open_vm_port')
        g.command('perform-maintenance', 'perform_maintenance', min_api='2017-03-30')
        g.command('redeploy', 'redeploy', supports_no_wait=True)
        g.custom_command('resize', 'resize_vm', supports_no_wait=True)
        g.command('restart', 'restart', supports_no_wait=True)
        g.custom_command('show', 'show_vm', table_transformer=transform_vm, exception_handler=empty_on_404)
        g.command('start', 'start', supports_no_wait=True)
        g.command('stop', 'power_off', supports_no_wait=True)
        g.generic_update_command('update', setter_name='update_vm', setter_type=compute_custom, supports_no_wait=True)
        g.generic_wait_command('wait', getter_name='get_instance_view', getter_type=compute_custom)

    with self.command_group('vm availability-set', compute_availset_sdk) as g:
        g.custom_command('convert', 'convert_av_set_to_managed_disk', min_api='2016-04-30-preview')
        g.custom_command('create', 'create_av_set', table_transformer=deployment_validate_table_format, supports_no_wait=True)
        g.command('delete', 'delete')
        g.command('list', 'list')
        g.command('list-sizes', 'list_available_sizes')
        g.command('show', 'get', exception_handler=empty_on_404)
        g.generic_update_command('update')

    with self.command_group('vm boot-diagnostics', compute_vm_sdk) as g:
        g.custom_command('disable', 'disable_boot_diagnostics')
        g.custom_command('enable', 'enable_boot_diagnostics')
        g.custom_command('get-boot-log', 'get_boot_log')

    with self.command_group('vm diagnostics', compute_vm_sdk) as g:
        g.custom_command('set', 'set_diagnostics_extension')
        g.custom_command('get-default-config', 'show_default_diagnostics_configuration')

    with self.command_group('vm disk', compute_vm_sdk, min_api='2017-03-30') as g:
        g.custom_command('attach', 'attach_managed_data_disk')
        g.custom_command('detach', 'detach_data_disk')

    with self.command_group('vm encryption', custom_command_type=compute_disk_encryption_custom) as g:
        g.custom_command('enable', 'encrypt_vm', validator=process_disk_encryption_namespace)
        g.custom_command('disable', 'decrypt_vm')
        g.custom_command('show', 'show_vm_encryption_status', exception_handler=empty_on_404)

    with self.command_group('vm extension', compute_vm_extension_sdk) as g:
        g.command('delete', 'delete')
        g.command('show', 'get', exception_handler=empty_on_404, table_transformer=transform_extension_show_table_output)
        g.custom_command('set', 'set_extension')
        g.custom_command('list', 'list_extensions', table_transformer='[].' + transform_extension_show_table_output)

    with self.command_group('vm extension image', compute_vm_extension_image_sdk) as g:
        g.command('show', 'get', exception_handler=empty_on_404)
        g.command('list-names', 'list_types')
        g.command('list-versions', 'list_versions')
        g.custom_command('list', 'list_vm_extension_images')

    with self.command_group('vm image', compute_vm_image_sdk) as g:
        g.command('list-offers', 'list_offers')
        g.command('list-publishers', 'list_publishers')
        g.command('list-skus', 'list_skus')
        g.custom_command('list', 'list_vm_images')
        g.custom_command('accept-terms', 'accept_market_ordering_terms')
        g.custom_command('show', 'show_vm_image', exception_handler=empty_on_404)

    with self.command_group('vm nic', compute_vm_sdk) as g:
        g.custom_command('add', 'add_vm_nic')
        g.custom_command('remove', 'remove_vm_nic')
        g.custom_command('set', 'set_vm_nic')
        g.custom_command('show', 'show_vm_nic', exception_handler=empty_on_404)
        g.custom_command('list', 'list_vm_nics')

    with self.command_group('vm run-command', compute_vm_run_sdk, operation_group='virtual_machine_run_commands') as g:
        g.custom_command('invoke', 'run_command_invoke')
        g.command('list', 'list')
        g.command('show', 'get')

    with self.command_group('vm secret', compute_vm_sdk) as g:
        g.custom_command('format', 'get_vm_format_secret', validator=process_vm_secret_format)
        g.custom_command('add', 'add_vm_secret')
        g.custom_command('list', 'list_vm_secrets')
        g.custom_command('remove', 'remove_vm_secret')

    with self.command_group('vm unmanaged-disk', compute_vm_sdk) as g:
        g.custom_command('attach', 'attach_unmanaged_data_disk')
        g.custom_command('detach', 'detach_data_disk')
        g.custom_command('list', 'list_unmanaged_disks')

    with self.command_group('vm user', compute_vm_sdk, supports_no_wait=True) as g:
        g.custom_command('update', 'set_user')
        g.custom_command('delete', 'delete_user')
        g.custom_command('reset-ssh', 'reset_linux_ssh')

    with self.command_group('vmss', compute_vmss_sdk, operation_group='virtual_machine_scale_sets') as g:
        g.custom_command('identity assign', 'assign_vmss_identity', validator=process_assign_identity_namespace)
        g.custom_command('identity remove', 'remove_vmss_identity', validator=process_remove_identity_namespace, min_api='2017-12-01')
        g.custom_command('identity show', 'show_vmss_identity')
        g.custom_command('create', 'create_vmss', transform=DeploymentOutputLongRunningOperation(self.cli_ctx, 'Starting vmss create'), supports_no_wait=True, table_transformer=deployment_validate_table_format, validator=process_vmss_create_namespace)
        g.custom_command('deallocate', 'deallocate_vmss', supports_no_wait=True)
        g.command('delete', 'delete', supports_no_wait=True)
        g.custom_command('delete-instances', 'delete_vmss_instances', supports_no_wait=True)
        g.custom_command('get-instance-view', 'get_vmss_instance_view', table_transformer='{ProvisioningState:statuses[0].displayStatus, PowerState:statuses[1].displayStatus}')
        g.custom_command('list', 'list_vmss', table_transformer=get_vmss_table_output_transformer(self))
        g.command('list-instances', 'list', command_type=compute_vmss_vm_sdk)
        g.custom_command('list-instance-connection-info', 'list_vmss_instance_connection_info')
        g.custom_command('list-instance-public-ips', 'list_vmss_instance_public_ips')
        g.command('list-skus', 'list_skus')
        g.custom_command('reimage', 'reimage_vmss', supports_no_wait=True, min_api='2017-03-30')
        g.command('perform-maintenance', 'perform_maintenance', min_api='2017-12-01')
        g.custom_command('restart', 'restart_vmss', supports_no_wait=True)
        g.custom_command('scale', 'scale_vmss', supports_no_wait=True)
        g.custom_command('show', 'show_vmss', exception_handler=empty_on_404, table_transformer=get_vmss_table_output_transformer(self, False))
        g.custom_command('start', 'start_vmss', supports_no_wait=True)
        g.custom_command('stop', 'stop_vmss', supports_no_wait=True)
        g.generic_update_command('update', getter_name='get_vmss', setter_name='set_vmss', supports_no_wait=True, command_type=compute_custom)
        g.custom_command('update-instances', 'update_vmss_instances', supports_no_wait=True)
        g.generic_wait_command('wait', getter_name='get_vmss', getter_type=compute_custom)

    with self.command_group('vmss diagnostics', compute_vmss_sdk) as g:
        g.custom_command('set', 'set_vmss_diagnostics_extension')
        g.custom_command('get-default-config', 'show_default_diagnostics_configuration')

    with self.command_group('vmss disk', compute_vmss_sdk, min_api='2017-03-30') as g:
        g.custom_command('attach', 'attach_managed_data_disk_to_vmss')
        g.custom_command('detach', 'detach_disk_from_vmss')

    with self.command_group('vmss encryption', custom_command_type=compute_disk_encryption_custom, min_api='2017-03-30') as g:
        g.custom_command('enable', 'encrypt_vmss', validator=process_disk_encryption_namespace)
        g.custom_command('disable', 'decrypt_vmss')
        g.custom_command('show', 'show_vmss_encryption_status', exception_handler=empty_on_404)

    with self.command_group('vmss extension', compute_vmss_sdk) as g:
        g.custom_command('delete', 'delete_vmss_extension')
        g.custom_command('show', 'get_vmss_extension', exception_handler=empty_on_404)
        g.custom_command('set', 'set_vmss_extension')
        g.custom_command('list', 'list_vmss_extensions')

    with self.command_group('vmss extension image', compute_vm_extension_image_sdk) as g:
        g.command('show', 'get', exception_handler=empty_on_404)
        g.command('list-names', 'list_types')
        g.command('list-versions', 'list_versions')
        g.custom_command('list', 'list_vm_extension_images')

    with self.command_group('vmss nic', network_nic_sdk) as g:
        g.command('list', 'list_virtual_machine_scale_set_network_interfaces')
        g.command('list-vm-nics', 'list_virtual_machine_scale_set_vm_network_interfaces')
        g.command('show', 'get_virtual_machine_scale_set_network_interface', exception_handler=empty_on_404)

    with self.command_group('vmss rolling-upgrade', compute_vmss_rolling_upgrade_sdk, min_api='2017-03-30') as g:
        g.command('cancel', 'cancel')
        g.command('get-latest', 'get_latest')
        g.command('start', 'start_os_upgrade')



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import json
import os

from knack.log import get_logger
from knack.util import CLIError

logger = get_logger(__name__)


MSI_LOCAL_ID = '[system]'


def get_target_network_api(cli_ctx):
    """ Since most compute calls don't need advanced network functionality, we can target a supported, but not
        necessarily latest, network API version is order to avoid having to re-record every test that uses VM create
        (which there are a lot) whenever NRP bumps their API version (which is often)!
    """
    from azure.cli.core.profiles import get_api_version, ResourceType
    version = get_api_version(cli_ctx, ResourceType.MGMT_NETWORK)
    if cli_ctx.cloud.profile == 'latest':
        version = '2018-01-01'
    return version


def read_content_if_is_file(string_or_file):
    content = string_or_file
    if os.path.exists(string_or_file):
        with open(string_or_file, 'r') as f:
            content = f.read()
    return content


def _resolve_api_version(cli_ctx, provider_namespace, resource_type, parent_path):
    from azure.cli.core.commands.client_factory import get_mgmt_service_client
    from azure.cli.core.profiles import ResourceType
    client = get_mgmt_service_client(cli_ctx, ResourceType.MGMT_RESOURCE_RESOURCES)
    provider = client.providers.get(provider_namespace)

    # If available, we will use parent resource's api-version
    resource_type_str = (parent_path.split('/')[0] if parent_path else resource_type)

    rt = [t for t in provider.resource_types  # pylint: disable=no-member
          if t.resource_type.lower() == resource_type_str.lower()]
    if not rt:
        raise CLIError('Resource type {} not found.'.format(resource_type_str))
    if len(rt) == 1 and rt[0].api_versions:
        npv = [v for v in rt[0].api_versions if 'preview' not in v.lower()]
        return npv[0] if npv else rt[0].api_versions[0]
    else:
        raise CLIError(
            'API version is required and could not be resolved for resource {}'
            .format(resource_type))


def log_pprint_template(template):
    logger.info('==== BEGIN TEMPLATE ====')
    logger.info(json.dumps(template, indent=2))
    logger.info('==== END TEMPLATE ====')


def check_existence(cli_ctx, value, resource_group, provider_namespace, resource_type,
                    parent_name=None, parent_type=None):
    # check for name or ID and set the type flags
    from azure.cli.core.commands.client_factory import get_mgmt_service_client
    from msrestazure.azure_exceptions import CloudError
    from msrestazure.tools import parse_resource_id
    from azure.cli.core.profiles import ResourceType
    resource_client = get_mgmt_service_client(cli_ctx, ResourceType.MGMT_RESOURCE_RESOURCES).resources

    id_parts = parse_resource_id(value)

    rg = id_parts.get('resource_group', resource_group)
    ns = id_parts.get('namespace', provider_namespace)

    if parent_name and parent_type:
        parent_path = '{}/{}'.format(parent_type, parent_name)
        resource_name = id_parts.get('child_name_1', value)
        resource_type = id_parts.get('child_type_1', resource_type)
    else:
        parent_path = ''
        resource_name = id_parts['name']
        resource_type = id_parts.get('type', resource_type)
    api_version = _resolve_api_version(cli_ctx, provider_namespace, resource_type, parent_path)

    try:
        resource_client.get(rg, ns, parent_path, resource_type, resource_name, api_version)
        return True
    except CloudError:
        return False


def create_keyvault_data_plane_client(cli_ctx):
    from azure.cli.core._profile import Profile

    def get_token(server, resource, scope):  # pylint: disable=unused-argument
        return Profile(cli_ctx=cli_ctx).get_login_credentials(resource)[0]._token_retriever()  # pylint: disable=protected-access

    from azure.keyvault import KeyVaultClient, KeyVaultAuthentication
    return KeyVaultClient(KeyVaultAuthentication(get_token))


def get_key_vault_base_url(cli_ctx, vault_name):
    suffix = cli_ctx.cloud.suffixes.keyvault_dns
    return 'https://{}{}'.format(vault_name, suffix)


def list_sku_info(cli_ctx, location=None):
    from ._client_factory import _compute_client_factory

    def _match_location(l, locations):
        return next((x for x in locations if x.lower() == l.lower()), None)

    client = _compute_client_factory(cli_ctx)
    result = client.resource_skus.list()
    if location:
        result = [r for r in result if _match_location(location, r.locations)]
    return result


def normalize_disk_info(image_data_disks=None, data_disk_sizes_gb=None, attach_data_disks=None, storage_sku=None,
                        os_disk_caching=None, data_disk_cachings=None, write_accelerator_settings=None):
    # we should return a dictionary with info like below and will emoit when see conflictions
    # {
    #   'os': { caching: 'Read', write_accelerator: None},
    #   0: { caching: 'None', write_accelerator: True},
    #   1: { caching: 'None', write_accelerator: True},
    # }
    from msrestazure.tools import is_valid_resource_id
    info = {}
    attach_data_disks = attach_data_disks or []
    image_data_disks = image_data_disks or []
    data_disk_sizes_gb = data_disk_sizes_gb or []
    info['os'] = {}

    for i in range(len(image_data_disks) + len(data_disk_sizes_gb) + len(attach_data_disks)):
        info[i] = {
            'lun': i
        }

    # fill in storage sku for managed data disks
    for i in range(len(image_data_disks) + len(data_disk_sizes_gb)):
        info[i]['managedDisk'] = {'storageAccountType': storage_sku}

    # fill in createOption
    for i in range(len(image_data_disks)):
        info[i]['createOption'] = 'fromImage'
    base = len(image_data_disks)
    for i in range(base, base + len(data_disk_sizes_gb)):
        info[i]['createOption'] = 'empty'
        info[i]['diskSizeGB'] = data_disk_sizes_gb[i]
    base = len(image_data_disks) + len(data_disk_sizes_gb)
    for i in range(base, base + len(attach_data_disks)):
        info[i]['createOption'] = 'attach'

    # fill in attached data disks details
    base = len(image_data_disks) + len(data_disk_sizes_gb)
    for i, d in enumerate(attach_data_disks):
        if is_valid_resource_id(d):
            info[base + i]['managedDisk'] = {'id': d}
        else:
            info[base + i]['vhd'] = {'uri': d}
            info[base + i]['name'] = d.split('/')[-1].split('.')[0]

    # fill in data disk caching
    if data_disk_cachings:
        update_disk_caching(info, data_disk_cachings)

    # fill in write accelerators
    if write_accelerator_settings:
        update_write_accelerator_settings(info, write_accelerator_settings)

    # default os disk caching to 'ReadWrite' unless set otherwise
    if os_disk_caching:
        info['os']['caching'] = os_disk_caching
    else:
        if info['os'].get('writeAcceleratorEnabled'):
            info['os']['caching'] = 'None'
        else:
            info['os']['caching'] = 'ReadWrite'

    return info


def update_disk_caching(model, caching_settings):

    def _update(model, lun, value):
        if isinstance(model, dict):
            luns = model.keys() if lun is None else [lun]
            for l in luns:
                if l not in model:
                    raise CLIError("data disk with lun of '{}' doesn't exist".format(lun))
                model[l]['caching'] = value
        else:
            if lun is None:
                disks = [model.os_disk] + (model.data_disks or [])
            elif lun == 'os':
                disks = [model.os_disk]
            else:
                disk = next((d for d in model.data_disks if d.lun == lun), None)
                if not disk:
                    raise CLIError("data disk with lun of '{}' doesn't exist".format(lun))
                disks = [disk]
            for disk in disks:
                disk.caching = value

    if len(caching_settings) == 1 and '=' not in caching_settings[0]:
        _update(model, None, caching_settings[0])
    else:
        for x in caching_settings:
            if '=' not in x:
                raise CLIError("usage error: please use 'LUN=VALUE' to configure caching on individual disk")
            lun, value = x.split('=', 1)
            lun = lun.lower()
            lun = int(lun) if lun != 'os' else lun
            _update(model, lun, value)


def update_write_accelerator_settings(model, write_accelerator_settings):

    def _update(model, lun, value):
        if isinstance(model, dict):
            luns = model.keys() if lun is None else [lun]
            for l in luns:
                if l not in model:
                    raise CLIError("data disk with lun of '{}' doesn't exist".format(lun))
                model[l]['writeAcceleratorEnabled'] = value
        else:
            if lun is None:
                disks = [model.os_disk] + (model.data_disks or [])
            elif lun == 'os':
                disks = [model.os_disk]
            else:
                disk = next((d for d in model.data_disks if d.lun == lun), None)
                if not disk:
                    raise CLIError("data disk with lun of '{}' doesn't exist".format(lun))
                disks = [disk]
            for disk in disks:
                disk.write_accelerator_enabled = value

    if len(write_accelerator_settings) == 1 and '=' not in write_accelerator_settings[0]:
        _update(model, None, write_accelerator_settings[0].lower() == 'true')
    else:
        for x in write_accelerator_settings:
            if '=' not in x:
                raise CLIError("usage error: please use 'LUN=VALUE' to configure write accelerator"
                               " on individual disk")
            lun, value = x.split('=', 1)
            lun = lun.lower()
            lun = int(lun) if lun != 'os' else lun
            _update(model, lun, value.lower() == 'true')
